/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <algorithm>
#include <memory>

#include "common/config.h"
#include "common/context.h"
#include "common/log.h"
#include "defs.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "recovery/log_manager.h"
#include "system/sm.h"
#include "transaction/txn_defs.h"

class DeleteExecutor : public AbstractExecutor {
 private:
  TabMeta tab_;                   // 表的元数据
  std::vector<Condition> conds_;  // delete的条件
  RmFileHandle *fh_;              // 表的数据文件句柄
  std::vector<Rid> rids_;         // 需要删除的记录的位置
  std::string tab_name_;          // 表名称
  SmManager *sm_manager_;
  std::unique_ptr<AbstractExecutor> scanner_;

 public:
  // 删除算子需要写操作，所以需要context来获取事务信息。
  DeleteExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Condition> conds,
                 std::unique_ptr<AbstractExecutor> scan, Context *context) {
    sm_manager_ = sm_manager;
    tab_name_ = tab_name;
    tab_ = sm_manager_->db_.get_table(tab_name);
    fh_ = sm_manager_->fhs_.at(tab_name).get();
    conds_ = conds;
    scanner_ = std::move(scan);
    AbstractExecutor::context_ = context;
  }

  // 删除算子的beginTuple() 不会被运行！！！
  void beginTuple() override {}

  std::unique_ptr<RmRecord> Next() override {
    ASSERT(context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd()),
           "get lock failed in delete.");

    open_conditions(conds_);

    for (scanner_->beginTuple(); !scanner_->is_end(); scanner_->nextTuple()) {
      rids_.push_back(scanner_->rid());
    }

    // 用于间隙锁判断
    std::vector<std::string> tbl_col_names;
    for (auto &col : tab_.cols) {
      tbl_col_names.push_back(col.name);
    }

    for (auto &rid : rids_) {
      auto rec = fh_->get_record(rid, context_);
      std::vector<Value> delete_values;
      for (auto &col : tab_.cols) {
        delete_values.push_back(get_value(rec.get(), col));
      }
      // // 对索引进行间隙锁判定
      // for (size_t i = 0; i < tab_.indexes.size(); ++i) {
      //   auto &index = tab_.indexes[i];
      //   auto ih = sm_manager_->ihs_
      //                 .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols))
      //                 .get();
      //   // 对于唯一索引下，如果没有违法唯一的前提下才会考虑间隙锁
      //   ASSERT(context_->lock_mgr_->lock_gap_on_index(context_->txn_, fh_->GetFd(), ih->get_fd(),
      //                                                 tbl_col_names, delete_values),
      //          "get lock failed in insert.");
      // }

      WriteRecord *d_rec = new WriteRecord(WType::DELETE_TUPLE, tab_name_, rid, *rec);
      context_->txn_->append_write_record(d_rec);

#ifdef RECOVERY_ON
      if (!context_->txn_->get_has_write()) {
        BeginLogRecord log_record(context_->txn_->get_transaction_id());
        log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
        lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
        context_->txn_->set_prev_lsn(prev_lsn);
        context_->txn_->get_has_write() = true;
      }
      DeleteLogRecord log_record(context_->txn_->get_transaction_id(), *rec, rid, tab_name_);
      log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
      lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
      context_->txn_->set_prev_lsn(prev_lsn);
#endif

      fh_->delete_record(rid, context_);

      for (size_t i = 0; i < tab_.indexes.size(); ++i) {
        auto &index = tab_.indexes[i];
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols))
                      .get();

        RmRecord key(index.col_tot_len);
        index.make_index_tuple(rec.get(), &key);

        ih->delete_entry(key.data, context_->txn_);
      }
    }

    return nullptr;
  }

  Rid &rid() override { return _abstract_rid; }
  bool is_end() const override { return true; }
};