/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <algorithm>
#include <cfloat>
#include <climits>
#include <cstddef>
#include <cstdint>
#include <memory>

#include "common/common.h"
#include "common/context.h"
#include "common/log.h"
#include "defs.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "index/ix_defs.h"
#include "index/ix_index_handle.h"
#include "index/ix_scan.h"
#include "record/rm_defs.h"
#include "system/sm.h"
#include "system/sm_meta.h"

class IndexScanExecutor : public AbstractExecutor {
 private:
  std::string tab_name_;              // 表名称
  TabMeta tab_;                       // 表的元数据
  std::vector<Condition> conds_;      // 扫描条件
  RmFileHandle *fh_;                  // 表的数据文件句柄
  std::vector<ColMeta> cols_;         // 需要读取的字段
  size_t len_;                        // 选取出来的一条记录的长度
  std::vector<Condition> fed_conds_;  // 扫描条件，和conds_字段相同

  std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
  IndexMeta index_meta_;                      // index scan涉及到的索引元数据
  std::vector<ColMeta>
      index_tuple_cols_;  // 记录了index中的tuple的信息。x相比index_meta.cols只修改了offset部分。

  Rid rid_;
  std::unique_ptr<RecScan> scan_;

  SmManager *sm_manager_;

 public:
  // 索引扫描算子需要读操作，所以需要context来获取事务信息。
  IndexScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds,
                    std::vector<std::string> index_col_names, Context *context) {
    sm_manager_ = sm_manager;
    tab_name_ = std::move(tab_name);
    tab_ = sm_manager_->db_.get_table(tab_name_);
    // index_no_ = index_no;
    index_col_names_ = index_col_names;

    AbstractExecutor::context_ = context;

    for (auto &index_meta : tab_.indexes) {
      size_t i;

      // 索引：index_col_names全部是索引cols的前缀。
      for (i = 0; i < index_col_names_.size(); ++i) {
        if (index_col_names_[i] != index_meta.cols[i].name) {
          break;
        }
      }

      if (i == index_col_names.size()) {
        index_meta_ = index_meta;
        break;
      }
    }

    fh_ = sm_manager_->fhs_.at(tab_name_).get();
    cols_ = tab_.cols;
    len_ = cols_.back().offset + cols_.back().len;

    std::map<CompOp, CompOp> swap_op = {
        {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT},
        {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
    };

    // 下面是把左边变成当前table。
    for (auto &cond : conds) {
      if (cond.lhs_col.tab_name != tab_name_) {
        // lhs is on other table, now rhs must be on this table
        assert(!cond.is_rhs_val && cond.rhs_col.tab_name == tab_name_);
        // swap lhs and rhs
        std::swap(cond.lhs_col, cond.rhs_col);
        cond.op = swap_op.at(cond.op);
      }

      // 现在条件的左表必然是tab_name_.
      // 现在左右表名不等，条件为等值，可以认为是归并连接。
      // 归并连接条件不可用作扫描算子的筛选。
      if (cond.op == OP_EQ && cond.is_rhs_column &&
          cond.lhs_col.tab_name != cond.rhs_col.tab_name) {
        ;  // 什么都不做。
      } else {
        fed_conds_.push_back(cond);  //
      }
    }

    conds_ = fed_conds_;

    int offset = 0;
    for (auto &index_meta : index_meta_.cols) {
      index_tuple_cols_.emplace_back(index_meta);
      index_tuple_cols_.back().offset = offset;
      offset += index_meta.len;
    }

#ifdef DEBUG
    std::cout << "index scan, start !!!" << std::endl;
    for (auto &col : cols_) {
      std::cout << col.name << ' ';
    }
    std::cout << std::endl;
#endif
  }

  void beginTuple() override {
    ASSERT(context_->lock_mgr_->lock_IS_on_table(context_->txn_, fh_->GetFd()),
           "get lock failed in index scan.");

    open_conditions(conds_);

    // 进行必要的类型转换。
    for (auto &cond : conds_) {
      if (cond.lhs_col.tab_name == tab_name_ && cond.is_rhs_val) {
        auto col_meta_iter =
            std::find_if(index_meta_.cols.begin(), index_meta_.cols.end(),
                         [&cond](ColMeta &col) { return cond.lhs_col.col_name == col.name; });
        if (col_meta_iter == index_meta_.cols.end()) {
          continue;
        }

        cond.rhs_val.convert_to_type(col_meta_iter->type);
        cond.rhs_val.resize(col_meta_iter->len);
      }
    }

    RmRecord key_start(index_meta_.col_tot_len);
    RmRecord key_end(index_meta_.col_tot_len);
    std::vector<const ColMeta *> lower_col;
    std::vector<const ColMeta *> upper_col;
    std::unordered_map<std::string, Value> left_, right_;

    auto ix_manager = sm_manager_->get_ix_manager();
    auto ix_index_handle =
        sm_manager_->ihs_.at(ix_manager->get_index_name(tab_name_, index_meta_.cols)).get();

    // 构造上下界。
    // int offset = 0; // offset是不需要的，因为ColMeta记录了offset的信息。
    for (auto &index_col_name : index_col_names_) {
      for (auto &condition : conds_) {
        if (condition.is_rhs_val && condition.lhs_col.tab_name == tab_name_ &&
            condition.lhs_col.col_name == index_col_name) {
          const char *rhs_key = condition.rhs_val.data();

          const ColMeta &index_col = *get_col(index_tuple_cols_, condition.lhs_col);

          // 下面逻辑存在问题，需要修改。
          // 改好了应该，但是可能还是存在某种错误。
          switch (condition.op) {
            case OP_EQ: {
              {
                auto iter = std::find_if(
                    lower_col.begin(), lower_col.end(),
                    [&index_col_name](const ColMeta *col) { return col->name == index_col_name; });

                if (iter == lower_col.end()) {
                  memcpy(key_start.data + index_col.offset, rhs_key, index_col.len);
                  lower_col.push_back(&index_col);
                  Value tmp = condition.rhs_val;
                  if (left_.find(index_col_name) == left_.end()) {
                    tmp.set_max();
                    left_[index_col_name] = tmp;
                  }
                  tmp = condition.rhs_val;
                  tmp.set_closed(condition.op);
                  tmp.min(left_[index_col_name]);
                  left_[index_col_name] = tmp;
                }
              }

              {
                auto iter = std::find_if(
                    upper_col.begin(), upper_col.end(),
                    [&index_col_name](const ColMeta *col) { return col->name == index_col_name; });

                if (iter == upper_col.end()) {
                  memcpy(key_end.data + index_col.offset, rhs_key, index_col.len);
                  upper_col.push_back(&index_col);
                  Value tmp = condition.rhs_val;
                  if (right_.find(index_col_name) == right_.end()) {
                    tmp.set_min();
                    left_[index_col_name] = tmp;
                  }
                  tmp = condition.rhs_val;
                  tmp.set_closed(condition.op);
                  tmp.max(left_[index_col_name]);
                  right_[index_col_name] = tmp;
                }
              }
              break;
            }
            // 下届 ：
            case OP_GT:
            case OP_GE: {
              auto iter = std::find_if(
                  lower_col.begin(), lower_col.end(),
                  [&index_col_name](const ColMeta *col) { return col->name == index_col_name; });

              if (iter == lower_col.end()) {
                memcpy(key_start.data + index_col.offset, rhs_key, index_col.len);
                lower_col.push_back(&index_col);
                Value tmp = condition.rhs_val;
                if (left_.find(index_col_name) == left_.end()) {
                  tmp.set_max();
                  left_[index_col_name] = tmp;
                }
                tmp = condition.rhs_val;
                tmp.set_closed(condition.op);
                tmp.min(left_[index_col_name]);
                left_[index_col_name] = tmp;
              }
              break;
            }
            case OP_LT:
            case OP_LE: {
              auto iter = std::find_if(
                  upper_col.begin(), upper_col.end(),
                  [&index_col_name](const ColMeta *col) { return col->name == index_col_name; });

              if (iter == upper_col.end()) {
                memcpy(key_end.data + index_col.offset, rhs_key, index_col.len);
                upper_col.push_back(&index_col);
                Value tmp = condition.rhs_val;
                if (right_.find(index_col_name) == right_.end()) {
                  tmp.set_min();
                  right_[index_col_name] = tmp;
                }
                tmp = condition.rhs_val;
                tmp.set_closed(condition.op);
                tmp.max(right_[index_col_name]);
                right_[index_col_name] = tmp;
              }
              break;
            }
            default:
              break;
          }
        }
      }
    }

    /* 添加间隙锁,*/
    ASSERT(context_->lock_mgr_->add_index_gap_lock(context_->txn_, fh_->GetFd(),
                                                   ix_index_handle->get_fd(), index_col_names_,
                                                   left_, right_),
           "get lock failed in index scan.");

    // 初始化 ix_san 所需要的参数
    Iid lower_Iid;  // 左闭右开，即使是 > 也包含左边界，只需要scan的时候check_conditions即可。
    Iid upper_Iid;

    // 需要考虑一下如果所给的上下界信息不完全，该怎么找。
    // 照理说应该lower找最小值,upper找最大值。不够的地方需要填充。
    // 如果空，或者第一个不是最左，那么从头开始匹配。
    if (lower_col.empty() || lower_col.front()->name != index_meta_.cols.front().name) {
      lower_Iid = ix_index_handle->leaf_begin();
    } else {
      if (lower_col.size() < index_tuple_cols_.size()) {
        // lower_col.size() < index_meta.cols 说明有些没有用上。

        // 双指针寻找。
        size_t i, j;
        for (i = 0, j = 0; i < lower_col.size(); ++j) {
          if (lower_col[i]->name == index_tuple_cols_[j].name) {
            ++i;
            continue;
          }

          // 不相等，说明这部分被漏掉了。
          switch (index_tuple_cols_[j].type) {
            case TYPE_INT: {
              int min_int = INT32_MIN;
              memcpy(key_start.data + index_tuple_cols_[j].offset, &min_int,
                     index_tuple_cols_[j].len);
            } break;
            case TYPE_FLOAT: {
              float min_float = FLT_MIN;
              memcpy(key_start.data + index_tuple_cols_[j].offset, &min_float,
                     index_tuple_cols_[j].len);
            } break;
            case TYPE_STRING: {
              memset(key_start.data + index_tuple_cols_[j].offset, 0, index_tuple_cols_[j].len);
            } break;
            default: {
              ;
            };
          }
        }

        for (; j < index_tuple_cols_.size(); ++j) {
          // 不相等，说明这部分被漏掉了。
          switch (index_tuple_cols_[j].type) {
            case TYPE_INT: {
              int min_int = INT32_MIN;
              memcpy(key_start.data + index_tuple_cols_[j].offset, &min_int,
                     index_tuple_cols_[j].len);
            } break;
            case TYPE_FLOAT: {
              float min_float = FLT_MIN;
              memcpy(key_start.data + index_tuple_cols_[j].offset, &min_float,
                     index_tuple_cols_[j].len);
            } break;
            case TYPE_STRING: {
              memset(key_start.data + index_tuple_cols_[j].offset, 0, index_tuple_cols_[j].len);
            } break;
            default: {
              ;
            };
          }
        }
      }

      lower_Iid = ix_index_handle->lower_bound(key_start.data);
    }

    if (upper_col.empty() || upper_col.front()->name != index_meta_.cols.front().name) {
      upper_Iid = ix_index_handle->leaf_end();
    } else {
      if (upper_col.size() < index_tuple_cols_.size()) {
        size_t i, j;
        for (i = 0, j = 0; i < upper_col.size(); ++j) {
          if (upper_col[i]->name == index_tuple_cols_[j].name) {
            ++i;
            continue;
          }

          // 不相等，说明这部分被漏掉了。
          switch (index_tuple_cols_[j].type) {
            case TYPE_INT: {
              int max_int = INT32_MAX;
              memcpy(key_end.data + index_tuple_cols_[j].offset, &max_int,
                     index_tuple_cols_[j].len);
            } break;
            case TYPE_FLOAT: {
              float max_float = FLT_MAX;
              memcpy(key_end.data + index_tuple_cols_[j].offset, &max_float,
                     index_tuple_cols_[j].len);
            } break;
            case TYPE_STRING: {
              memset(key_end.data + index_tuple_cols_[j].offset, CHAR_MAX,
                     index_tuple_cols_[j].len);
            } break;
            default: {
              ;
            };
          }
        }

        for (; j < index_tuple_cols_.size(); ++j) {
          switch (index_tuple_cols_[j].type) {
            case TYPE_INT: {
              int max_int = INT32_MAX;
              memcpy(key_end.data + index_tuple_cols_[j].offset, &max_int,
                     index_tuple_cols_[j].len);
            } break;
            case TYPE_FLOAT: {
              float max_float = FLT_MAX;
              memcpy(key_end.data + index_tuple_cols_[j].offset, &max_float,
                     index_tuple_cols_[j].len);
            } break;
            case TYPE_STRING: {
              memset(key_end.data + index_tuple_cols_[j].offset, CHAR_MAX,
                     index_tuple_cols_[j].len);
            } break;
            default: {
              ;
            } break;
          }
        }
      }

      upper_Iid = ix_index_handle->upper_bound(key_end.data);
    }

    // 初始化 索引表扫描操作接口
    scan_ = std::make_unique<IxScan>(ix_index_handle, lower_Iid, upper_Iid, sm_manager_->get_bpm());

    while (!scan_->is_end()) {
      auto rcd = fh_->get_record(scan_->rid(), context_);
      ASSERT(context_->lock_mgr_->lock_shared_on_record(context_->txn_, rid_, fh_->GetFd()),
             "get rec lock failed in select.");

      if (!check_conditions(rcd.get(), rcd.get(), conds_, cols_, cols_)) {
        scan_->next();
      } else {
        rid_ = scan_->rid();
        // 读取的记录放行级读锁

        break;
      }
    }
  }
  // nextTuple 和beginTuple 获取到san 之后的操作都和顺序遍历一样
  void nextTuple() override {
    for (scan_->next(); !scan_->is_end(); scan_->next()) {
      auto rcd = fh_->get_record(scan_->rid(), context_);
      ASSERT(context_->lock_mgr_->lock_shared_on_record(context_->txn_, rid_, fh_->GetFd()),
             "get rec lock failed in select.");

      if (check_conditions(rcd.get(), rcd.get(), conds_, cols_, cols_)) {
        rid_ = scan_->rid();
        // 读取的记录放行级读锁

        break;
      }
    }
  }

  std::unique_ptr<RmRecord> Next() override { return fh_->get_record(rid_, context_); }

  Rid &rid() override { return rid_; }

  bool is_end() const override { return scan_->is_end(); }

  const std::vector<ColMeta> &cols() const override { return cols_; }

  size_t tupleLen() const override { return len_; };
};