/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstddef>

#include "common/config.h"
#include "common/context.h"
#include "common/log.h"
#include "errors.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "recovery/log_manager.h"
#include "system/sm.h"
#include "transaction/transaction.h"
#include "transaction/txn_defs.h"

class InsertExecutor : public AbstractExecutor {
 private:
  TabMeta tab_;                // 表的元数据
  std::vector<Value> values_;  // 需要插入的数据
  RmFileHandle *fh_;           // 表的数据文件句柄
  std::string tab_name_;       // 表名称
  Rid rid_;  // 插入的位置，由于系统默认插入时不指定位置，因此当前rid_在插入后才赋值
  SmManager *sm_manager_;

 public:
  // 插入算子需要写操作，所以需要context来获取事务信息。
  InsertExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Value> values,
                 Context *context) {
    sm_manager_ = sm_manager;
    tab_ = sm_manager_->db_.get_table(tab_name);
    values_ = values;
    tab_name_ = tab_name;
    if (values.size() != tab_.cols.size()) {
      throw InvalidValueCountError();
    }
    fh_ = sm_manager_->fhs_.at(tab_name).get();
    AbstractExecutor::context_ = context;
  };

  // 插入算子的beginTuple() 不会被运行！！！
  void beginTuple() override {}

  std::unique_ptr<RmRecord> Next() override {
    // Make record buffer
    RmRecord rec(fh_->get_file_hdr().record_size);
    std::vector<std::string> col_names;
    std::vector<Value> insert_values;
    // 因为插入暂时只支持一个值还是很好保持原子性的。
    for (size_t i = 0; i < values_.size(); i++) {
      auto &col = tab_.cols[i];
      auto &val = values_[i];
      col_names.push_back(tab_.cols[i].name);
      val.convert_to_type(col.type);
      val.resize(col.len);
      //  val.init_raw(col.len);
      insert_values.push_back(values_[i]);
      memcpy(rec.data + col.offset, val.data(), col.len);
    }

    // 如果表没有索引直接上表排斥锁,test
    if (tab_.indexes.empty()) {
      ASSERT(context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd()),
             "get lock failed in insert.");
    } else {
      ASSERT(context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd()),
             "get lock failed in insert.");
      // 对索引进行间隙锁判定
      for (size_t i = 0; i < tab_.indexes.size(); ++i) {
        auto &index = tab_.indexes[i];
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols))
                      .get();
        // 对于唯一索引下，如果没有违法唯一的前提下才会考虑间隙锁
        ASSERT(context_->lock_mgr_->lock_gap_on_index(context_->txn_, fh_->GetFd(), ih->get_fd(),
                                                      col_names, insert_values),
               "get lock failed in insert.");
      }
    }

    rid_ = fh_->insert_record(rec.data, context_);

    if (!tab_.indexes.empty()) {
      // 对于插入的记录放置行级别写锁
      ASSERT(context_->lock_mgr_->lock_exclusive_on_record(context_->txn_, rid_, fh_->GetFd()),
             "get rec lock failed in insert.");
    }

#ifdef RECOVERY_ON
    {
      if (!context_->txn_->get_has_write()) {
        BeginLogRecord log_record(context_->txn_->get_transaction_id());
        log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
        lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
        context_->txn_->set_prev_lsn(prev_lsn);
        context_->txn_->get_has_write() = true;
      }
      InsertLogRecord log_record(context_->txn_->get_transaction_id(), rec, rid_, tab_name_);
      log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
      lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
      context_->txn_->set_prev_lsn(prev_lsn);
    }
#endif

    // Insert into index
    // 插入时，如果违反唯一性索引，需要抛出异常。
    size_t i;
    for (i = 0; i < tab_.indexes.size(); ++i) {
      auto &index = tab_.indexes[i];
      auto ih =
          sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols))
              .get();
      // char *key = new char[index.col_tot_len]; // 这种也不进行垃圾回收的垃圾写法就别拿出来了。
      RmRecord key(index.col_tot_len);
      index.make_index_tuple(&rec, &key);

      std::vector<Rid> rids;
      ih->get_value(key.data, &rids, context_->txn_);

      // 其中某个索引违反了唯一索引。
      if (!rids.empty()) {
        break;
      }
      // // 对于唯一索引下，如果没有违法唯一的前提下才会考虑间隙锁
      // ASSERT(context_->lock_mgr_->lock_gap_on_index(context_->txn_, fh_->GetFd(), ih->get_fd(),
      //                                               col_names, insert_values),
      //        "get lock failed in insert.");

      ih->insert_entry(key.data, rid_, context_->txn_);
    }

    // 不需要进行回复。
    if (i == tab_.indexes.size()) {
      // 这里才算是插入成功了，所以这里才把写记录写入。
      // 但是如果设计故障恢复怎么办？？？
      // 添加一个标记
      context_->txn_->set_gap_insert();
      WriteRecord *w_rec = new WriteRecord(WType::INSERT_TUPLE, tab_name_, rid_);
      context_->txn_->append_write_record(w_rec);
      // Insert into record file
      return nullptr;
    }

    // 不需要了，现在改成先插入index中再插入
    // 进行恢复：之前插入的全部进行删除.
    // fh_->delete_record(rid_, context_);  // 表中删除

    // 对索引进行删除恢复。
    for (size_t j = 0; j < i; ++j) {  // 最后一个索引也就是 第i个索引 没有插入，所以不需要进行删除。
      auto &index = tab_.indexes[j];
      auto ih =
          sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols))
              .get();
      // char *key = new char[index.col_tot_len]; // 这种也不进行垃圾回收的垃圾写法就别拿出来了。
      RmRecord key(index.col_tot_len);
      index.make_index_tuple(&rec, &key);

      ih->delete_entry(key.data, context_->txn_);
    }

#ifdef RECOVERY_ON
    {
      if (!context_->txn_->get_has_write()) {
        BeginLogRecord log_record(context_->txn_->get_transaction_id());
        log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
        lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
        context_->txn_->set_prev_lsn(prev_lsn);
        context_->txn_->get_has_write() = true;
      }
      DeleteLogRecord log_record(context_->txn_->get_transaction_id(), rec, rid_, tab_name_);
      log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
      lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
      context_->txn_->set_prev_lsn(prev_lsn);
    }
#endif

    fh_->delete_record(rid_, context_);

    throw InternalError("insert_entry has exists");
    return nullptr;
  }
  Rid &rid() override { return rid_; }
  bool is_end() const override { return true; }
};