/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstddef>
#include <memory>

#include "common/common.h"
#include "common/log.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
 private:
  std::unique_ptr<AbstractExecutor> left_;   // 左儿子节点（需要join的表）
  std::unique_ptr<AbstractExecutor> right_;  // 右儿子节点（需要join的表）
  size_t len_;                               // join后获得的每条记录的长度
  std::vector<ColMeta> cols_;                // join后获得的记录的字段

  std::vector<Condition> fed_conds_;        // join条件
  std::map<TabCol, Value> prev_feed_dict_;  // 字段到值的映射
  std::unique_ptr<RmRecord> current_rec_;
  std::unique_ptr<RmRecord> left_rec_;  // 把左边的tuple存下来可以少调用很多次left_->Next();
  // size_t right_i_;
  // std::vector<std::unique_ptr<RmRecord>> right_recs_;

 public:
  // 嵌套链接算子不需要读写操作，所以不需要context来获取事务信息。
  NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left,
                         std::unique_ptr<AbstractExecutor> right, std::vector<Condition> conds) {
    left_ = std::move(left);
    right_ = std::move(right);
    len_ = left_->tupleLen() + right_->tupleLen();
    cols_ = left_->cols();
    auto right_cols = right_->cols();
    for (auto &col : right_cols) {
      col.offset += left_->tupleLen();
    }

    cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
    fed_conds_ = std::move(conds);
  }

  void beginTuple() override {
    open_conditions(fed_conds_);

    left_->beginTuple();
    left_rec_ = left_->Next();
    if (left_->is_end()) {
      return;
    }

    right_->beginTuple();

    // right_i_ = 0;
    make_join();
  }

  void nextTuple() override {
    if (is_end()) {
      return;
    }

    right_->nextTuple();
    // ++right_i_;
    if (right_->is_end()) {
      left_->nextTuple();
      left_rec_ = left_->Next();
      right_->beginTuple();
      // right_i_ = 0;
    }
    make_join();
  }

  // 这里Next不应该move，不然如果多次调用Next的话，后面会出现错误.
  // std::unique_ptr<RmRecord> Next() override { return std::move(current_rec_); }

  // 把拷贝的一份返回出去。
  std::unique_ptr<RmRecord> Next() override { return std::make_unique<RmRecord>(*current_rec_); }

  bool is_end() const override { return left_->is_end(); }

  Rid &rid() override { return _abstract_rid; }

  const std::vector<ColMeta> &cols() const override { return cols_; };

  void make_join() {
    while (true) {  // 因为每次调用Next都会消耗资源（会重新分配RmRecord）所以尽量少调用Next
      std::unique_ptr<RmRecord> right_rec = right_->Next();  //
      // auto &right_rec = right_recs_[right_i_];

      if (check_conditions(left_rec_.get(), right_rec.get(), fed_conds_, left_->cols(),
                           right_->cols())) {
        make_tuple(left_rec_, right_rec);
        break;
      } else {
        right_->nextTuple();
        // ++right_i_;

        if (right_->is_end()) {
          left_->nextTuple();
          left_rec_ = left_->Next();
          right_->beginTuple();
          // right_i_ = 0;
        }

        if (left_->is_end()) {
          break;
        }
      }
    }
  }

  size_t tupleLen() const override { return len_; }

  void make_tuple(std::unique_ptr<RmRecord> &left_tuple, std::unique_ptr<RmRecord> &right_tuple) {
    current_rec_ = std::make_unique<RmRecord>(len_);
    size_t left_len = left_->tupleLen();
    memcpy(current_rec_->data, left_tuple->data, left_len);
    memcpy(current_rec_->data + left_len, right_tuple->data, right_->tupleLen());

#ifdef DEBUG
    printf("left_tuple : ");
    for (auto &left_col : left_->cols()) {
      Value val;
      val.type = left_col.type;
      val.set_data(left_tuple->data + left_col.offset, left_col.len);

      printf("%s ", val.to_string().c_str());
    }

    printf(" right_tuple : ");

    for (auto &right_col : right_->cols()) {
      Value val;
      val.type = right_col.type;
      val.set_data(right_tuple->data + right_col.offset, right_col.len);

      printf("%s ", val.to_string().c_str());
    }

    printf(" tuple : ");
    for (auto &col : cols_) {
      Value val;
      val.type = col.type;
      val.set_data(current_rec_->data + col.offset, col.len);

      printf("%s ", val.to_string().c_str());
    }
    printf("\n");
#endif
  }
};