/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstddef>
#include <vector>

#include "common/common.h"
#include "common/config.h"
#include "errors.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "system/sm.h"
#include "system/sm_meta.h"

class ProjectionExecutor : public AbstractExecutor {
 private:
  std::unique_ptr<AbstractExecutor> prev_;  // 投影节点的儿子节点
  std::vector<ColMeta> cols_;               // 需要投影的字段
  size_t len_;                              // 字段总长度

 public:
 // 投影算子不需要读写操作，所以不需要context来获取事务信息。
  ProjectionExecutor(std::unique_ptr<AbstractExecutor> prev, const std::vector<TabCol> &sel_cols) {
    prev_ = std::move(prev);

    size_t curr_offset = 0;
    auto &prev_cols = prev_->cols();

    for (auto &sel_col : sel_cols) {
      auto iter = std::find_if(prev_cols.begin(), prev_cols.end(), [&sel_col](const ColMeta &col) {
        return sel_col.tab_name == col.tab_name && sel_col.col_name == col.name &&
               sel_col.aggregation_type == col.aggregation_type;
      });

      ColMeta tmp = *iter;
      tmp.offset = curr_offset;
      cols_.push_back(tmp);

      curr_offset += iter->len;
    }

    len_ = curr_offset;
  }

  void beginTuple() override { prev_->beginTuple(); }

  void nextTuple() override { prev_->nextTuple(); }

  std::unique_ptr<RmRecord> Next() override {
    auto prev_tuple = prev_->Next();
    auto tuple = std::make_unique<RmRecord>(len_);
    auto &prev_metas = prev_->cols();

    for (auto &col_meta : cols_) {
      auto child_iter =
          std::find_if(prev_metas.begin(), prev_metas.end(), [&col_meta](const ColMeta &col) {
            return col.tab_name == col_meta.tab_name && col.name == col_meta.name &&
                   col.aggregation_type == col_meta.aggregation_type;
          });

      memcpy(tuple->data + col_meta.offset, prev_tuple->data + child_iter->offset, col_meta.len);
    }

    return tuple;
  }

  Rid &rid() override { return _abstract_rid; }
  const std::vector<ColMeta> &cols() const override { return cols_; }
  bool is_end() const override { return prev_->is_end(); }
  size_t tupleLen() const override { return len_; }
};