/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <unistd.h>

#include <cstddef>
#include <memory>
#include <queue>
#include <string>

#include "common/common.h"
#include "common/log.h"
#include "execution/executor_index_scan.h"
#include "execution/executor_seq_scan.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "record_printer.h"
#include "system/sm.h"

enum class Table_scn_type { LeftTbl, RightTbl, SeqScan, IndexScan };

struct PairComparator {
  bool operator()(const std::pair<RmRecord, int> &left,
                  const std::pair<RmRecord, int> &right) const {
    return left.second > right.second;  // 注意这里的 '>'，因为我们使用的是 std::greater<>
  }
};

class SortMergeJoinExecutor : public AbstractExecutor {
 private:
  std::unique_ptr<AbstractExecutor> left_;            // 左儿子节点（需要join的表）
  std::unique_ptr<AbstractExecutor> right_;           // 右儿子节点（需要join的表）
  size_t len_;                                        // join后获得的每条记录的长度
  std::vector<ColMeta> cols_, cols_left, cols_right;  // join后获得的记录的字段,cols_left 左表
  ColMeta col_set_left, col_set_right;                // 用于比较的左、右表字段

  std::vector<Condition> fed_conds_;        // join条件
  std::map<TabCol, Value> prev_feed_dict_;  // 字段到值的映射
  std::unique_ptr<RmRecord> current_rec_;
  std::unique_ptr<RmRecord> left_rec_;  // 把左边的tuple存下来可以少调用很多次left_->Next();
  std::vector<std::unique_ptr<RmRecord>> tuples_left, tuples_right;  // 左右边全部数据，用于排序
  std::vector<std::unique_ptr<RmRecord>>::iterator current_tuple_left, current_tuple_right;
  bool need_reverse_;  // 是否需要调整中间排序的结果顺序
  bool need_sleep_;
  // 外部排序
  size_t memSize = 2000;
  std::string target_sort_file_name_ = "sorted_results.txt";
  std::string left_file_name_ = "sorted_results_left.txt";
  std::string right_file_name_ = "sorted_results_right.txt";

  std::unique_ptr<AbstractExecutor> root_sort;
  ColMeta col_set_sort;
  std::vector<ColMeta> cols_sort;
  std::ifstream left_read, right_read;
  bool left_read_end = false, right_read_end = false;
  RmRecord *current_left, *current_right;
  std::unique_ptr<RmRecord> ptr_record_left, ptr_record_right;  // 用于管理指针

 public:
  SortMergeJoinExecutor(std::unique_ptr<AbstractExecutor> left,
                        std::unique_ptr<AbstractExecutor> right, std::vector<Condition> conds,
                        Context *context, bool need_reverse, bool need_sleep = false) {
    need_sleep_ = need_sleep;
    left_ = std::move(left);
    right_ = std::move(right);
    need_reverse_ = need_reverse;
    len_ = left_->tupleLen() + right_->tupleLen();
    cols_ = left_->cols();
    cols_left = left_->cols();
    cols_right = right_->cols();
    auto right_cols = right_->cols();

    for (auto &col : right_cols) {
      col.offset += left_->tupleLen();
    }

    cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
    fed_conds_ = std::move(conds);

    col_set_left = *get_col(cols_left, (fed_conds_[0].lhs_col.tab_name == cols_left[0].tab_name)
                                           ? fed_conds_[0].lhs_col
                                           : fed_conds_[0].rhs_col);

    col_set_right = *get_col(cols_right, (fed_conds_[0].rhs_col.tab_name == cols_right[0].tab_name)
                                             ? fed_conds_[0].rhs_col
                                             : fed_conds_[0].lhs_col);
    AbstractExecutor::context_ = context;
  }
  // 检测连接的表的扫描算子的类型
  Table_scn_type table_san_type(Table_scn_type table_scan) {
    SeqScanExecutor *derived1Ptr;
    IndexScanExecutor *derived2Ptr;
    switch (table_scan) {
      case Table_scn_type::LeftTbl:
        derived1Ptr = dynamic_cast<SeqScanExecutor *>(left_.get());
        derived2Ptr = dynamic_cast<IndexScanExecutor *>(left_.get());
        break;
      case Table_scn_type::RightTbl:
        derived1Ptr = dynamic_cast<SeqScanExecutor *>(right_.get());
        derived2Ptr = dynamic_cast<IndexScanExecutor *>(right_.get());
        break;
      default:
        break;
    }
    // 根据转换结果判断类型
    if (derived1Ptr) {
      return Table_scn_type::SeqScan;
    } else if (derived2Ptr) {
      return Table_scn_type::IndexScan;
    }
    return Table_scn_type::SeqScan;
  }

  // 通过从文件中读取一条记录,引用的方式
  void get_cmp_rmd_from_file(Cmp_Record &record, std::string &line, size_t &n,
                             std::vector<Value> &values, int &fileIndex) {
    std::stringstream str_stream(line);
    for (size_t i = 0; i < n; ++i) {
      std::getline(str_stream, line, '|');
      line = trim(line);  // 去除前后空格
      line >> values[i];
      values[i].resize(cols_sort[i].len);
      memcpy(record.data + cols_sort[i].offset, values[i].data(), cols_sort[i].len);
    }
    // 用于优先队列比较
    record.compare_tye = col_set_sort.type;
    record.compare_offset = col_set_sort.offset;
    record.compare_len = col_set_sort.len;
    record.compare_index = fileIndex;
  }

  // 归并k个排序文件
  void kWayMerge(const std::vector<std::string> &files, const std::string &outputFile,
                 std::vector<std::string> captions) {
    std::fstream outfile, outfile_target;
    outfile.open(outputFile, std::ios::out | std::ios::app);
    outfile_target.open(target_sort_file_name_, std::ios::out | std::ios::app);
    std::priority_queue<Cmp_Record, std::vector<Cmp_Record>, std::greater<Cmp_Record>> minHeap,
        minHeap_tmp;

    // 从每个文件读取第一个元素
    std::vector<std::ifstream> ifStreams;
    size_t n = cols_sort.size();  // 记录列数量
    std::vector<Value> values(n);

    for (size_t i = 0; i < n; ++i) {
      values[i].type = cols_sort[i].type;
    }
    int fileIndex = 0;  // 添加一个计数器
    Cmp_Record record(root_sort->tupleLen());
    for (const auto &file : files) {
      std::ifstream ifStream(file);
      ifStreams.emplace_back(std::move(ifStream));
      std::string line;

      if (std::getline(ifStreams.back(), line)) {
        get_cmp_rmd_from_file(record, line, n, values, fileIndex);
        // std::stringstream str_stream(line);
        // for (size_t i = 0; i < n; ++i) {
        //   std::getline(str_stream, line, '|');
        //   line = trim(line);  // 去除前后空格
        //   line >> values[i];
        //   values[i].resize(cols_sort[i].len);
        //   memcpy(record.data + cols_sort[i].offset, values[i].data(), cols_sort[i].len);
        // }
        // // 用于优先队列比较
        // record.compare_tye = col_set_sort.type;
        // record.compare_offset = col_set_sort.offset;
        // record.compare_len = col_set_sort.len;
        // record.compare_index = fileIndex;
        minHeap.emplace(record);  // files.size() is a placeholder, we will update it later
      }
      fileIndex++;
    }
    while (!minHeap.empty()) {
      auto tmp = minHeap.top();
      minHeap.pop();
      tmp.compare_index =
          &ifStreams[tmp.compare_index] - &ifStreams[0];  // update index to point to correct
      minHeap_tmp.emplace(tmp);
    }
    minHeap = minHeap_tmp;

    // 表头
    outfile_target << "|";
    for (size_t i = 0; i < captions.size(); ++i) {
      outfile_target << " " << captions[i] << " |";
    }
    outfile_target << "\n";

    while (!minHeap.empty()) {
      auto tuple = minHeap.top();
      minHeap.pop();
      std::vector<std::string> columns;
      for (auto &col : cols_sort) {
        std::string col_str_2;
        char *rec_buf = tuple.data + col.offset;
        if (col.type == TYPE_INT) {
          col_str_2 = std::to_string(*(int *)rec_buf);
        } else if (col.type == TYPE_FLOAT) {
          col_str_2 = std::to_string(*(float *)rec_buf);
        } else if (col.type == TYPE_STRING) {
          col_str_2 = std::string((char *)rec_buf, col.len);
          col_str_2.resize(strlen(col_str_2.c_str()));
        }
        columns.push_back(col_str_2);
      }
      outfile_target << "|";
      for (size_t i = 0; i < columns.size(); ++i) {
        outfile << " " << columns[i] << " |";
        outfile_target << " " << columns[i] << " |";
      }
      outfile << "\n";
      outfile_target << "\n";
      std::string nextLine;
      if (std::getline(ifStreams[tuple.compare_index], nextLine)) {
        // 检查流的状态，确保没有错误发生
        if (ifStreams[tuple.compare_index].fail()) {
          // 处理错误情况，比如打印错误信息或跳过此文件流
          std::cerr << "Error reading from file stream: " << tuple.compare_index << std::endl;
        }
        Cmp_Record record(root_sort->tupleLen());
        get_cmp_rmd_from_file(record, nextLine, n, values, tuple.compare_index);
        // std::stringstream str_stream(nextLine);
        // for (size_t i = 0; i < n; ++i) {
        //   std::getline(str_stream, nextLine, '|');
        //   nextLine = trim(nextLine);  // 去除前后空格
        //   nextLine >> values[i];

        //   values[i].resize(cols_sort[i].len);
        //   memcpy(record.data + cols_sort[i].offset, values[i].data(), cols_sort[i].len);
        // }
        // // 用于优先队列比较
        // record.compare_tye = col_set_sort.type;
        // record.compare_offset = col_set_sort.offset;
        // record.compare_len = col_set_sort.len;
        // record.compare_index = tuple.compare_index;
        minHeap.emplace(record);
      }
    }
    outfile.close();
    outfile_target.close();
  }
  // 二路外部排序的主要函数
  void externalSort(const std::string &outputFilename, size_t memSize,
                    std::vector<std::string> captions) {
    if (outputFilename == left_file_name_) {
      root_sort = std::move(left_);
      col_set_sort = col_set_left;
      cols_sort = cols_left;
    } else {
      root_sort = std::move(right_);
      col_set_sort = col_set_right;
      cols_sort = cols_right;
    }
    std::string line;
    std::vector<std::string> tempFiles;
    size_t chunkId = 0;

    // 读取并排序数据块
    for (root_sort->beginTuple(); !root_sort->is_end(); root_sort->nextTuple()) {
      auto tuple = root_sort->Next();
      std::vector<std::unique_ptr<RmRecord>> chunk;
      chunk.emplace_back(std::move(tuple));
      while (chunk.size() < memSize && (!root_sort->is_end())) {
        root_sort->nextTuple();
        if (root_sort->is_end()) break;
        tuple = root_sort->Next();
        chunk.emplace_back(std::move(tuple));
      }

      // 排序
      Value lhs;
      Value rhs;
      lhs.type = col_set_sort.type;
      rhs.type = col_set_sort.type;
      auto cmp = [this, &lhs, &rhs](std::unique_ptr<RmRecord> &left,
                                    std::unique_ptr<RmRecord> &right) {
        lhs.set_data(left->data + col_set_sort.offset, col_set_sort.len);
        rhs.set_data(right->data + col_set_sort.offset, col_set_sort.len);

        CompResult res = lhs.compare_value(rhs);
        return res == CompResult::LT;
      };
      std::sort(chunk.begin(), chunk.end(), cmp);
      // 写入临时文件
      txn_id_t t_id = context_->txn_->get_transaction_id();
      std::string tempFile = "temp_" + std::to_string(t_id) + std::to_string(chunkId++);
      std::ofstream outFile;
      outFile.open(tempFile, std::ios::out | std::ios::app);
      // 左表表头
      // outFile << "|";
      // for (size_t i = 0; i < captions.size(); ++i) {
      //   outFile << " " << captions[i] << " |";
      // }
      // outFile << "\n";
      for (auto &tuple : chunk) {
        std::vector<std::string> columns;
        for (auto &col : cols_sort) {
          std::string col_str_2;
          char *rec_buf = tuple->data + col.offset;
          if (col.type == TYPE_INT) {
            col_str_2 = std::to_string(*(int *)rec_buf);
          } else if (col.type == TYPE_FLOAT) {
            col_str_2 = std::to_string(*(float *)rec_buf);
          } else if (col.type == TYPE_STRING) {
            col_str_2 = std::string((char *)rec_buf, col.len);
            col_str_2.resize(strlen(col_str_2.c_str()));
          }
          columns.push_back(col_str_2);
        }
        //   outFile << "|";
        for (size_t i = 0; i < columns.size(); ++i) {
          outFile << columns[i] << " |";
        }
        outFile << "\n";
      }
      tempFiles.push_back(tempFile);
      if (root_sort->is_end()) break;
    }

    // 多路归并
    kWayMerge(tempFiles, outputFilename, captions);

    // 清理临时文件
    for (const auto &file : tempFiles) {
      remove(file.c_str());
    }
    // 复原
    if (outputFilename == left_file_name_) {
      left_ = std::move(root_sort);
    } else {
      right_ = std::move(root_sort);
    }
  }

  // 获取元素每列数据转为为字符串
  void set_col_from_tuple(std::unique_ptr<RmRecord> &tuple, std::vector<ColMeta> &cols,
                          std::vector<std::string> &columns) {
    std::string col_str;
    for (auto &col : cols) {
      char *rec_buf = tuple->data + col.offset;
      if (col.type == TYPE_INT) {
        col_str = std::to_string(*(int *)rec_buf);
      } else if (col.type == TYPE_FLOAT) {
        col_str = std::to_string(*(float *)rec_buf);
      } else if (col.type == TYPE_STRING) {
        col_str = std::string((char *)rec_buf, col.len);
        col_str.resize(strlen(col_str.c_str()));
      }
      columns.push_back(col_str);
    }
  }

  void beginTuple() override {
    open_conditions(fed_conds_);
    // 表头
    std::vector<std::string> captions, captions_right;
    captions.reserve(cols_left.size());
    for (auto &sel_col : cols_left) {
      captions.push_back(sel_col.name);
    }
    captions_right.reserve(cols_right.size());
    for (auto &sel_col : cols_right) {
      captions_right.push_back(sel_col.name);
    }
    std::fstream outfile_target, outfile;
    if (need_reverse_ == true) {
      if (table_san_type(Table_scn_type::LeftTbl) == Table_scn_type::SeqScan) {
        externalSort(left_file_name_, memSize, captions);
      } else {
        outfile_target.open(target_sort_file_name_, std::ios::out | std::ios::app);
        outfile.open(left_file_name_, std::ios::out | std::ios::app);
        // 左表表头
        outfile_target << "|";
        for (size_t i = 0; i < captions.size(); ++i) {
          outfile_target << " " << captions[i] << " |";
        }
        outfile_target << "\n";
        for (left_->beginTuple(); !left_->is_end(); left_->nextTuple()) {
          auto tuple = left_->Next();
          std::vector<std::string> columns;
          set_col_from_tuple(tuple, cols_left, columns);
          // for (auto &col : cols_left) {
          //   std::string col_str_2;
          //   char *rec_buf = tuple->data + col.offset;
          //   if (col.type == TYPE_INT) {
          //     col_str_2 = std::to_string(*(int *)rec_buf);
          //   } else if (col.type == TYPE_FLOAT) {
          //     col_str_2 = std::to_string(*(float *)rec_buf);
          //   } else if (col.type == TYPE_STRING) {
          //     col_str_2 = std::string((char *)rec_buf, col.len);
          //     col_str_2.resize(strlen(col_str_2.c_str()));
          //   }
          //   columns.push_back(col_str_2);
          // }
          outfile_target << "|";
          for (size_t i = 0; i < columns.size(); ++i) {
            outfile_target << " " << columns[i] << " |";
            outfile << " " << columns[i] << " |";
          }
          outfile_target << "\n";
          outfile << "\n";
        }
        outfile_target.close();
        outfile.close();
      }
      if (table_san_type(Table_scn_type::RightTbl) == Table_scn_type::SeqScan) {
        externalSort(right_file_name_, memSize, captions_right);
      } else {
        outfile_target.open(target_sort_file_name_, std::ios::out | std::ios::app);
        outfile.open(right_file_name_, std::ios::out | std::ios::app);
        // 右表表头
        outfile_target << "|";
        for (size_t i = 0; i < captions_right.size(); ++i) {
          outfile_target << " " << captions_right[i] << " |";
        }
        outfile_target << "\n";
        for (right_->beginTuple(); !right_->is_end(); right_->nextTuple()) {
          auto tuple = right_->Next();
          std::vector<std::string> columns;
          set_col_from_tuple(tuple, cols_right, columns);
          // for (auto &col : cols_right) {
          //   std::string col_str;
          //   char *rec_buf = tuple->data + col.offset;
          //   if (col.type == TYPE_INT) {
          //     col_str = std::to_string(*(int *)rec_buf);
          //   } else if (col.type == TYPE_FLOAT) {
          //     col_str = std::to_string(*(float *)rec_buf);
          //   } else if (col.type == TYPE_STRING) {
          //     col_str = std::string((char *)rec_buf, col.len);
          //     col_str.resize(strlen(col_str.c_str()));
          //   }
          //   columns.push_back(col_str);
          // }
          outfile_target << "|";
          for (size_t i = 0; i < columns.size(); ++i) {
            outfile_target << " " << columns[i] << " |";
            outfile << " " << columns[i] << " |";
          }
          outfile_target << "\n";
          outfile << "\n";
        }
        outfile_target.close();
        outfile.close();
      }
    } else {
      if (table_san_type(Table_scn_type::RightTbl) == Table_scn_type::SeqScan) {
        externalSort(right_file_name_, 2, captions_right);
      } else {
        outfile_target.open(target_sort_file_name_, std::ios::out | std::ios::app);
        outfile.open(right_file_name_, std::ios::out | std::ios::app);
        // 右表表头
        outfile_target << "|";
        for (size_t i = 0; i < captions_right.size(); ++i) {
          outfile_target << " " << captions_right[i] << " |";
        }
        outfile_target << "\n";
        for (right_->beginTuple(); !right_->is_end(); right_->nextTuple()) {
          auto tuple = right_->Next();
          std::vector<std::string> columns;
          set_col_from_tuple(tuple, cols_right, columns);
          // for (auto &col : cols_right) {
          //   std::string col_str;
          //   char *rec_buf = tuple->data + col.offset;
          //   if (col.type == TYPE_INT) {
          //     col_str = std::to_string(*(int *)rec_buf);
          //   } else if (col.type == TYPE_FLOAT) {
          //     col_str = std::to_string(*(float *)rec_buf);
          //   } else if (col.type == TYPE_STRING) {
          //     col_str = std::string((char *)rec_buf, col.len);
          //     col_str.resize(strlen(col_str.c_str()));
          //   }
          //   columns.push_back(col_str);
          // }
          outfile_target << "|";
          for (size_t i = 0; i < columns.size(); ++i) {
            outfile_target << " " << columns[i] << " |";
            outfile << " " << columns[i] << " |";
          }
          outfile_target << "\n";
          outfile << "\n";
        }
        outfile_target.close();
        outfile.close();
      }
      if (table_san_type(Table_scn_type::LeftTbl) == Table_scn_type::SeqScan) {
        externalSort(left_file_name_, 2, captions);
      } else {
        outfile_target.open(target_sort_file_name_, std::ios::out | std::ios::app);
        outfile.open(left_file_name_, std::ios::out | std::ios::app);
        // 左表表头
        outfile_target << "|";
        for (size_t i = 0; i < captions.size(); ++i) {
          outfile_target << " " << captions[i] << " |";
        }
        outfile_target << "\n";
        for (left_->beginTuple(); !left_->is_end(); left_->nextTuple()) {
          auto tuple = left_->Next();
          std::vector<std::string> columns;
          set_col_from_tuple(tuple, cols_left, columns);
          // for (auto &col : cols_left) {
          //   std::string col_str_2;
          //   char *rec_buf = tuple->data + col.offset;
          //   if (col.type == TYPE_INT) {
          //     col_str_2 = std::to_string(*(int *)rec_buf);
          //   } else if (col.type == TYPE_FLOAT) {
          //     col_str_2 = std::to_string(*(float *)rec_buf);
          //   } else if (col.type == TYPE_STRING) {
          //     col_str_2 = std::string((char *)rec_buf, col.len);
          //     col_str_2.resize(strlen(col_str_2.c_str()));
          //   }
          //   columns.push_back(col_str_2);
          // }
          outfile_target << "|";
          for (size_t i = 0; i < columns.size(); ++i) {
            outfile_target << " " << columns[i] << " |";
            outfile << " " << columns[i] << " |";
          }
          outfile_target << "\n";
          outfile << "\n";
        }
        outfile_target.close();
        outfile.close();
      }
    }
    // externalSort(left_file_name_, 2, captions);
    // RecordPrinter rec_printer(cols_left.size()), rec_printer_right(cols_right.size());
    // // print header into file
    // std::fstream outfile;
    // outfile.open("sorted_results.txt", std::ios::out | std::ios::app);
    // for (left_->beginTuple(); !left_->is_end(); left_->nextTuple()) {
    //   auto tuple = left_->Next();
    //   tuples_left.emplace_back(std::move(tuple));
    // }

    // if (table_san_type(Table_scn_type::LeftTbl) == Table_scn_type::SeqScan) {
    //   Value lhs;
    //   Value rhs;
    //   lhs.type = col_set_left.type;
    //   rhs.type = col_set_left.type;
    //   auto cmp = [this, &lhs, &rhs](std::unique_ptr<RmRecord> &left,
    //                                 std::unique_ptr<RmRecord> &right) {
    //     lhs.set_data(left->data + col_set_left.offset, col_set_left.len);
    //     rhs.set_data(right->data + col_set_left.offset, col_set_left.len);

    //     CompResult res = lhs.compare_value(rhs);
    //     return res == CompResult::LT;
    //   };
    //   std::sort(tuples_left.begin(), tuples_left.end(), cmp);
    // }
    // // 右表
    // captions_right.reserve(cols_right.size());
    // for (auto &sel_col : cols_right) {
    //   captions_right.push_back(sel_col.name);
    // }
    // for (right_->beginTuple(); !right_->is_end(); right_->nextTuple()) {
    //   auto tuple = right_->Next();
    //   tuples_right.emplace_back(std::move(tuple));
    // }
    // if (table_san_type(Table_scn_type::RightTbl) == Table_scn_type::SeqScan) {
    //   Value lhs_right;
    //   Value rhs_right;
    //   lhs_right.type = col_set_right.type;
    //   rhs_right.type = col_set_right.type;
    //   auto cmp_right = [this, &lhs_right, &rhs_right](std::unique_ptr<RmRecord> &left,
    //                                                   std::unique_ptr<RmRecord> &right) {
    //     lhs_right.set_data(left->data + col_set_right.offset, col_set_right.len);
    //     rhs_right.set_data(right->data + col_set_right.offset, col_set_right.len);

    //     CompResult res = lhs_right.compare_value(rhs_right);
    //     return res == CompResult::LT;
    //   };
    //   std::sort(tuples_right.begin(), tuples_right.end(), cmp_right);
    // }

    // if (need_reverse_ == true) {
    //   // 左表表头
    //   outfile << "|";
    //   for (size_t i = 0; i < captions.size(); ++i) {
    //     outfile << " " << captions[i] << " |";
    //   }
    //   outfile << "\n";
    //   for (auto &tuple : tuples_left) {
    //     std::vector<std::string> columns;
    //     for (auto &col : cols_left) {
    //       std::string col_str_2;
    //       char *rec_buf = tuple->data + col.offset;
    //       if (col.type == TYPE_INT) {
    //         col_str_2 = std::to_string(*(int *)rec_buf);
    //       } else if (col.type == TYPE_FLOAT) {
    //         col_str_2 = std::to_string(*(float *)rec_buf);
    //       } else if (col.type == TYPE_STRING) {
    //         col_str_2 = std::string((char *)rec_buf, col.len);
    //         col_str_2.resize(strlen(col_str_2.c_str()));
    //       }
    //       columns.push_back(col_str_2);
    //     }
    //     outfile << "|";
    //     for (size_t i = 0; i < columns.size(); ++i) {
    //       outfile << " " << columns[i] << " |";
    //     }
    //     outfile << "\n";
    //   }
    //   // 右表表头
    //   outfile << "|";
    //   for (size_t i = 0; i < captions_right.size(); ++i) {
    //     outfile << " " << captions_right[i] << " |";
    //   }
    //   outfile << "\n";
    //   for (auto &tuple : tuples_right) {
    //     std::vector<std::string> columns;
    //     for (auto &col : cols_right) {
    //       std::string col_str;
    //       char *rec_buf = tuple->data + col.offset;
    //       if (col.type == TYPE_INT) {
    //         col_str = std::to_string(*(int *)rec_buf);
    //       } else if (col.type == TYPE_FLOAT) {
    //         col_str = std::to_string(*(float *)rec_buf);
    //       } else if (col.type == TYPE_STRING) {
    //         col_str = std::string((char *)rec_buf, col.len);
    //         col_str.resize(strlen(col_str.c_str()));
    //       }
    //       columns.push_back(col_str);
    //     }
    //     outfile << "|";
    //     for (size_t i = 0; i < columns.size(); ++i) {
    //       outfile << " " << columns[i] << " |";
    //     }
    //     outfile << "\n";
    //   }
    // } else {
    //   // 右表表头
    //   outfile << "|";
    //   for (size_t i = 0; i < captions_right.size(); ++i) {
    //     outfile << " " << captions_right[i] << " |";
    //   }
    //   outfile << "\n";
    //   for (auto &tuple : tuples_right) {
    //     std::vector<std::string> columns;
    //     for (auto &col : cols_right) {
    //       std::string col_str;
    //       char *rec_buf = tuple->data + col.offset;
    //       if (col.type == TYPE_INT) {
    //         col_str = std::to_string(*(int *)rec_buf);
    //       } else if (col.type == TYPE_FLOAT) {
    //         col_str = std::to_string(*(float *)rec_buf);
    //       } else if (col.type == TYPE_STRING) {
    //         col_str = std::string((char *)rec_buf, col.len);
    //         col_str.resize(strlen(col_str.c_str()));
    //       }
    //       columns.push_back(col_str);
    //     }
    //     outfile << "|";
    //     for (size_t i = 0; i < columns.size(); ++i) {
    //       outfile << " " << columns[i] << " |";
    //     }
    //     outfile << "\n";
    //   }
    //   // 左表表头
    //   outfile << "|";
    //   for (size_t i = 0; i < captions.size(); ++i) {
    //     outfile << " " << captions[i] << " |";
    //   }
    //   outfile << "\n";
    //   for (auto &tuple : tuples_left) {
    //     std::vector<std::string> columns;
    //     for (auto &col : cols_left) {
    //       std::string col_str_2;
    //       char *rec_buf = tuple->data + col.offset;
    //       if (col.type == TYPE_INT) {
    //         col_str_2 = std::to_string(*(int *)rec_buf);
    //       } else if (col.type == TYPE_FLOAT) {
    //         col_str_2 = std::to_string(*(float *)rec_buf);
    //       } else if (col.type == TYPE_STRING) {
    //         col_str_2 = std::string((char *)rec_buf, col.len);
    //         col_str_2.resize(strlen(col_str_2.c_str()));
    //       }
    //       columns.push_back(col_str_2);
    //     }
    //     outfile << "|";
    //     for (size_t i = 0; i < columns.size(); ++i) {
    //       outfile << " " << columns[i] << " |";
    //     }
    //     outfile << "\n";
    //   }
    // }
    // outfile.close();
    // current_tuple_left = tuples_left.begin();
    // current_tuple_right = tuples_right.begin();
    left_read.open(left_file_name_);
    right_read.open(right_file_name_);
    get_left_tuple();
    get_right_tuple();
    make_join();
  }

  void nextTuple() override {
    Value lhs, rhs;
    lhs.type = col_set_left.type;
    rhs.type = col_set_right.type;
    current_left = ptr_record_left.get();
    current_right = ptr_record_right.get();
    lhs.set_data(current_left->data + col_set_left.offset, col_set_left.len);
    rhs.set_data(current_right->data + col_set_right.offset, col_set_right.len);
    CompResult res = lhs.compare_value(rhs);
    if (res == CompResult::LT) {
      if (left_read_end == false) get_left_tuple(), current_left = ptr_record_left.get();
    } else {
      if (right_read_end == false) get_right_tuple(), current_right = ptr_record_right.get();
    }
    if (left_read_end == true || right_read_end == true) {
      left_read_end = true;
      right_read_end = true;
      return;
    }
    make_join();
  }

  // 这里Next不应该move，不然如果多次调用Next的话，后面会出现错误.
  // std::unique_ptr<RmRecord> Next() override { return std::move(current_rec_); }

  // 把拷贝的一份返回出去。
  std::unique_ptr<RmRecord> Next() override { return std::make_unique<RmRecord>(*current_rec_); }
  // bool is_end_test() { return current_tuple_left == tuples_left.end(); }
  bool is_end() const override {  // return current_tuple_left == tuples_left.end();
    // bool is_left_end = current_tuple_left == tuples_left.end();
    // bool is_right_end = current_tuple_right == tuples_right.end();
    // return is_left_end || is_right_end;
    return left_read_end == true || right_read_end == true;
  }

  Rid &rid() override { return _abstract_rid; }

  const std::vector<ColMeta> &cols() const override { return cols_; };

  // void make_join_test() {
  //   while (true) {  // 因为每次调用Next都会消耗资源（会重新分配RmRecord）所以尽量少调用Next
  //     //  std::unique_ptr<RmRecord> right_rec =
  //     std::make_unique<RmRecord>(**current_tuple_right); RmRecord *left =
  //     current_tuple_left->get();    // 直接获取RmRecord指针 RmRecord *right =
  //     current_tuple_right->get();  // 直接获取RmRecord指针 if (check_conditions(left, right,
  //     fed_conds_, left_->cols(), right_->cols())) {
  //       current_rec_ = std::make_unique<RmRecord>(len_);
  //       size_t left_len = left_->tupleLen();
  //       memcpy(current_rec_->data, left->data, left_len);
  //       memcpy(current_rec_->data + left_len, right->data, right_->tupleLen());
  //       break;
  //     } else {
  //       ++current_tuple_right;
  //       if (current_tuple_right == tuples_right.end()) {
  //         ++current_tuple_left;
  //         //  left_rec_ = std::make_unique<RmRecord>(**current_tuple_left);
  //         current_tuple_right = tuples_right.begin();
  //       }
  //       if (is_end_test()) {
  //         break;
  //       }
  //     }
  //   }
  // }
  // 简单的 trim 函数实现
  std::string trim(const std::string &str) {
    size_t first = str.find_first_not_of(' ');
    if (std::string::npos == first) {
      return str;  // 如果没有找到非空格字符，则返回空字符串
    }

    size_t last = str.find_last_not_of(' ');
    return str.substr(first, (last - first + 1));
  }

  void get_left_tuple() {
    std::string line;
    size_t n = cols_left.size();
    std::vector<Value> values(n);
    for (size_t i = 0; i < n; ++i) {
      values[i].type = cols_left[i].type;
    }
    // RmRecord record(left_->tupleLen());
    // RmRecord *record = new RmRecord(left_->tupleLen());
    std::unique_ptr<RmRecord> record = std::make_unique<RmRecord>(left_->tupleLen());
    if (std::getline(left_read, line)) {
      std::stringstream str_stream(line);
      for (size_t i = 0; i < n; ++i) {
        std::getline(str_stream, line, '|');
        line = trim(line);  // 去除前后空格
        line >> values[i];

        values[i].resize(cols_left[i].len);
        memcpy(record->data + cols_left[i].offset, values[i].data(), cols_left[i].len);
      }
    } else {
      left_read_end = true;
      left_read.close();
    }
    ptr_record_left = std::move(record);  // 将所有权转移到 ptr_record_left
    // ptr_record_left = std::make_unique<RmRecord>(record);
    // std::unique_ptr<RmRecord> ptr(record);
    //   return ptr;
    // return record;
    return;
  }
  void get_right_tuple() {
    std::string line;
    size_t n = cols_right.size();
    std::vector<Value> values(n);
    for (size_t i = 0; i < n; ++i) {
      values[i].type = cols_right[i].type;
    }
    // RmRecord record(right_->tupleLen());
    //   RmRecord *record = new RmRecord(right_->tupleLen());
    std::unique_ptr<RmRecord> record = std::make_unique<RmRecord>(right_->tupleLen());
    if (std::getline(right_read, line)) {
      std::stringstream str_stream(line);
      for (size_t i = 0; i < n; ++i) {
        std::getline(str_stream, line, '|');
        line = trim(line);  // 去除前后空格
        line >> values[i];
        values[i].resize(cols_right[i].len);
        memcpy(record->data + cols_right[i].offset, values[i].data(), cols_right[i].len);
      }
    } else {
      right_read_end = true;
      right_read.close();
    }
    ptr_record_right = std::move(record);  // 将所有权转移到 ptr_record_right
    // std::unique_ptr<RmRecord> ptr(record);
    // return ptr;
    return;
    // return record;
  }
  void make_join() {
    while (true) {  // 因为每次调用Next都会消耗资源（会重新分配RmRecord）所以尽量少调用Next
      //  RmRecord *left = current_tuple_left->get();  // 直接获取RmRecord指针
      // auto left_rec = get_left_tuple();
      //  get_left_tuple();
      // std::unique_ptr<RmRecord> ptr_record_left(get_left_tuple());
      current_left = ptr_record_left.get();  // 直接获取RmRecord指针

      //   auto right_rec = get_right_tuple();
      // get_right_tuple();
      current_right = ptr_record_right.get();  // 直接获取RmRecord指针
      if (check_conditions(current_left, current_right, fed_conds_, left_->cols(),
                           right_->cols())) {
        current_rec_ = std::make_unique<RmRecord>(len_);
        size_t left_len = left_->tupleLen();
        memcpy(current_rec_->data, current_left->data, left_len);
        memcpy(current_rec_->data + left_len, current_right->data, right_->tupleLen());
        break;
      } else {
        // 移动指针，准备检查下一个记录对
        Value lhs, rhs;
        lhs.type = col_set_left.type;
        rhs.type = col_set_right.type;
        lhs.set_data(current_left->data + col_set_left.offset, col_set_left.len);
        rhs.set_data(current_right->data + col_set_right.offset, col_set_right.len);
        CompResult res = lhs.compare_value(rhs);
        if (res == CompResult::LT) {
          if (left_read_end == false) get_left_tuple(), current_left = ptr_record_left.get();
        } else {
          if (right_read_end == false) get_right_tuple(), current_right = ptr_record_right.get();
        }
        if (left_read_end == true || right_read_end == true) break;
      }
    }
  }

  size_t tupleLen() const override { return len_; }

  void make_tuple(std::unique_ptr<RmRecord> &left_tuple, std::unique_ptr<RmRecord> &right_tuple) {
    current_rec_ = std::make_unique<RmRecord>(len_);
    size_t left_len = left_->tupleLen();
    memcpy(current_rec_->data, left_tuple->data, left_len);
    memcpy(current_rec_->data + left_len, right_tuple->data, right_->tupleLen());

#ifdef DEBUG
    printf("left_tuple : ");
    for (auto &left_col : left_->cols()) {
      Value val;
      val.type = left_col.type;
      val.set_data(left_tuple->data + left_col.offset, left_col.len);

      printf("%s ", val.to_string().c_str());
    }

    printf(" right_tuple : ");

    for (auto &right_col : right_->cols()) {
      Value val;
      val.type = right_col.type;
      val.set_data(right_tuple->data + right_col.offset, right_col.len);

      printf("%s ", val.to_string().c_str());
    }

    printf(" tuple : ");
    for (auto &col : cols_) {
      Value val;
      val.type = col.type;
      val.set_data(current_rec_->data + col.offset, col.len);

      printf("%s ", val.to_string().c_str());
    }
    printf("\n");
#endif
  }
};