/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstddef>
#include <memory>

#include "common/config.h"
#include "common/context.h"
#include "common/log.h"
#include "defs.h"
#include "errors.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "index/ix_index_handle.h"
#include "record/rm_defs.h"
#include "recovery/log_manager.h"
#include "system/sm.h"
#include "system/sm_meta.h"
#include "transaction/txn_defs.h"

class UpdateExecutor : public AbstractExecutor {
 private:
  TabMeta tab_;
  std::vector<Condition> conds_;
  RmFileHandle *fh_;
  std::vector<Rid> rids_;
  std::string tab_name_;
  std::vector<SetClause> set_clauses_;
  std::unique_ptr<AbstractExecutor> scanner_;
  SmManager *sm_manager_;

 public:
  // 更新算子需要写操作，所以需要context来获取事务信息。
  UpdateExecutor(SmManager *sm_manager, const std::string &tab_name,
                 std::vector<SetClause> set_clauses, std::vector<Condition> conds,
                 std::unique_ptr<AbstractExecutor> scanner, Context *context) {
    sm_manager_ = sm_manager;
    tab_name_ = tab_name;
    set_clauses_ = set_clauses;
    tab_ = sm_manager_->db_.get_table(tab_name);
    fh_ = sm_manager_->fhs_.at(tab_name).get();
    conds_ = conds;
    scanner_ = std::move(scanner);

    AbstractExecutor::context_ = context;
  }

  // 更新算子的beginTuple() 不会被运行！！！
  void beginTuple() override {}

  // 当更新的值违反了唯一性索引，需要抛出异常。
  // 更新麻烦的点在于，如果被更新的tuple不止一个，那么在恢复的时候很难办，所以不得不保存旧的tuples。
  std::unique_ptr<RmRecord> Next() override {
    // 来，上锁。
    ASSERT(context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd()),
           "get lock failed in update.");

    for (scanner_->beginTuple(); !scanner_->is_end(); scanner_->nextTuple()) {
      rids_.push_back(scanner_->rid());
    }

    std::vector<std::unique_ptr<RmRecord>> old_recs;  // 保存旧的tuples。

    // 先从索引中全部删除, 但是表中没有删除。
    for (auto &rid : rids_) {
      auto old_rec = fh_->get_record(rid, context_);

      for (size_t i = 0; i < tab_.indexes.size(); ++i) {
        auto &ix_meta = tab_.indexes[i];
        auto ix = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, ix_meta.cols))
                      .get();

        RmRecord key(ix_meta.col_tot_len);
        ix_meta.make_index_tuple(old_rec.get(), &key);
        ix->delete_entry(key.data, context_->txn_);
      }

      old_recs.emplace_back(std::move(old_rec));
    }

    size_t rid_i;
    for (rid_i = 0; rid_i < rids_.size(); ++rid_i) {
      auto &rid = rids_[rid_i];
      // 新份tuple，更新之后进行修改。
      auto &old_rec = old_recs[rid_i];
      auto new_rec = std::make_unique<RmRecord>(*old_rec);

      // 构建新值。
      for (auto &set_clause : set_clauses_) {
        auto col_meta_iter = std::find_if(
            tab_.cols.begin(), tab_.cols.end(),
            [&set_clause](ColMeta &col) { return col.name == set_clause.lhs.col_name; });

        if (col_meta_iter == tab_.cols.end()) {
          throw InternalError("The field not exists in the table.");
        }

        if (set_clause.op == OP_UB) {
          set_clause.rhs.convert_to_type(col_meta_iter->type);
          set_clause.rhs.resize(col_meta_iter->len);
          memcpy(new_rec->data + col_meta_iter->offset, set_clause.rhs.data(), col_meta_iter->len);
        } else {
          auto rhs_col_meta_iter = std::find_if(
              tab_.cols.begin(), tab_.cols.end(),
              [&set_clause](ColMeta &col) { return col.name == set_clause.rhs_col.col_name; });

          if (rhs_col_meta_iter == tab_.cols.end()) {
            throw InternalError("The field not exists in the table.");
          }

          Value val;
          val.type = rhs_col_meta_iter->type;
          val.set_data(old_rec->data + rhs_col_meta_iter->offset, rhs_col_meta_iter->len);
          val.cal_op(set_clause.rhs, set_clause.op);
          val.convert_to_type(col_meta_iter->type);
          val.resize(col_meta_iter->len);
          memcpy(new_rec->data + col_meta_iter->offset, val.data(), col_meta_iter->len);
        }
      }

#ifdef RECOVERY_ON
      {
        if (!context_->txn_->get_has_write()) {
          BeginLogRecord log_record(context_->txn_->get_transaction_id());
          log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
          lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
          context_->txn_->set_prev_lsn(prev_lsn);
          context_->txn_->get_has_write() = true;
        }
        UpdateLogRecord log_record(context_->txn_->get_transaction_id(), *old_rec, *new_rec, rid,
                                   tab_name_);

        log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
        lsn_t prev_lsn_ = context_->log_mgr_->add_log_to_buffer(&log_record);
        context_->txn_->set_prev_lsn(prev_lsn_);
      }
#endif

      fh_->update_record(rid, new_rec->data, context_);

      // 插入。
      size_t i;
      for (i = 0; i < tab_.indexes.size(); ++i) {
        auto &index_meta = tab_.indexes[i];
        auto index =
            sm_manager_->ihs_
                .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols))
                .get();

        RmRecord update_rec(index_meta.col_tot_len);
        index_meta.make_index_tuple(new_rec.get(), &update_rec);

        std::vector<Rid> rids;
        index->get_value(update_rec.data, &rids, context_->txn_);

        if (!rids.empty()) {  // 新值已经存在，违反唯一性原则，需要恢复并抛出异常。
          break;
        }

        index->insert_entry(update_rec.data, rid, context_->txn_);
      }

      // 成功插入。
      if (i == tab_.indexes.size()) {
        continue;
      }

#ifdef RECOVERY_ON
      {
        if (!context_->txn_->get_has_write()) {
          BeginLogRecord log_record(context_->txn_->get_transaction_id());
          log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
          lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
          context_->txn_->set_prev_lsn(prev_lsn);
          context_->txn_->get_has_write() = true;
        }
        UpdateLogRecord log_record(context_->txn_->get_transaction_id(), *new_rec, *old_rec, rid,
                                   tab_name_);

        log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
        lsn_t prev_lsn_ = context_->log_mgr_->add_log_to_buffer(&log_record);
        context_->txn_->set_prev_lsn(prev_lsn_);
      }
#endif

      // 第 i 个改回来。
      fh_->update_record(rid, old_rec->data, context_);

      // 前面 i 个删去新值。
      for (size_t j = 0; j < i; ++j) {  // 第i个索引没有更新，
        auto &index_meta = tab_.indexes[j];
        auto index =
            sm_manager_->ihs_
                .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols))
                .get();

        RmRecord update_rec(index_meta.col_tot_len);
        index_meta.make_index_tuple(new_rec.get(), &update_rec);

        // 删除更新值，插入老值。
        index->delete_entry(update_rec.data, context_->txn_);
      }

      // 存在违反索引唯一性的情况出现。相当于undo了吧。
      // 统一删除。
      for (size_t j = 0; j < rid_i; ++j) {
        auto &rid = rids_[j];
        auto &old_rec = old_recs[j];
        auto new_rec = std::make_unique<RmRecord>(*old_rec);

        // 构建新值。
        for (auto &set_clause : set_clauses_) {
          auto col_meta_iter = std::find_if(
              tab_.cols.begin(), tab_.cols.end(),
              [&set_clause](ColMeta &col) { return col.name == set_clause.lhs.col_name; });

          if (col_meta_iter == tab_.cols.end()) {
            throw InternalError("The field not exists in the table.");
          }

          if (set_clause.op == OP_UB) {
            set_clause.rhs.convert_to_type(col_meta_iter->type);
            set_clause.rhs.resize(col_meta_iter->len);
            memcpy(new_rec->data + col_meta_iter->offset, set_clause.rhs.data(),
                   col_meta_iter->len);
          } else {
            auto rhs_col_meta_iter = std::find_if(
                tab_.cols.begin(), tab_.cols.end(),
                [&set_clause](ColMeta &col) { return col.name == set_clause.rhs_col.col_name; });

            if (rhs_col_meta_iter == tab_.cols.end()) {
              throw InternalError("The field not exists in the table.");
            }

            Value val;
            val.type = rhs_col_meta_iter->type;
            val.set_data(old_rec->data + rhs_col_meta_iter->offset, rhs_col_meta_iter->len);
            val.cal_op(set_clause.rhs, set_clause.op);
            val.convert_to_type(col_meta_iter->type);
            val.resize(col_meta_iter->len);
            memcpy(new_rec->data + col_meta_iter->offset, val.data(), col_meta_iter->len);
          }
        }

#ifdef RECOVERY_ON
        if (!context_->txn_->get_has_write()) {
          BeginLogRecord log_record(context_->txn_->get_transaction_id());
          log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
          lsn_t prev_lsn = context_->log_mgr_->add_log_to_buffer(&log_record);
          context_->txn_->set_prev_lsn(prev_lsn);
          context_->txn_->get_has_write() = true;
        }
        UpdateLogRecord log_record(context_->txn_->get_transaction_id(), *new_rec, *old_rec, rid,
                                   tab_name_);
        log_record.prev_lsn_ = context_->txn_->get_prev_lsn();
        lsn_t prev_lsn_ = context_->log_mgr_->add_log_to_buffer(&log_record);
        context_->txn_->set_prev_lsn(prev_lsn_);
#endif

        fh_->update_record(rid, old_rec->data, context_);  // 表恢复成原来的tuple。

        for (auto &index_meta : tab_.indexes) {
          auto index =
              sm_manager_->ihs_
                  .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols))
                  .get();

          RmRecord update_rec(index_meta.col_tot_len);
          index_meta.make_index_tuple(new_rec.get(), &update_rec);

          // 删除更新值，插入老值。
          index->delete_entry(update_rec.data, context_->txn_);
        }
      }

      for (size_t j = 0; j < rids_.size(); ++j) {
        auto &rid = rids_[j];
        auto &old_rec = old_recs[j];

        for (auto &index_meta : tab_.indexes) {
          auto index =
              sm_manager_->ihs_
                  .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols))
                  .get();

          RmRecord rec(index_meta.col_tot_len);
          index_meta.make_index_tuple(old_rec.get(), &rec);

          // 删除更新值，插入老值。
          index->insert_entry(rec.data, rid, context_->txn_);
        }
      }
      break;
    }

    // 所有的tuple都被正常的更新了。
    if (rid_i == rids_.size()) {
      for (size_t i = 0; i < rids_.size(); ++i) {
        auto &rec = old_recs[i];
        auto &rid = rids_[i];

        WriteRecord *u_rec = new WriteRecord(WType::UPDATE_TUPLE, tab_name_, rid, *rec);
        context_->txn_->append_write_record(u_rec);
      }
      return nullptr;
    }

    throw InternalError("updated value has exists in index.");
    return nullptr;
  }

  Rid &rid() override { return _abstract_rid; }
  bool is_end() const override { return true; }
};