/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cstddef>
#include <memory>

#include "common/common.h"
#include "defs.h"
#include "ix_defs.h"
#include "transaction/transaction.h"

enum class Operation { FIND = 0, INSERT, DELETE };  // 三种操作：查找、插入、删除

static const bool binary_search = false;

// 这个函数没用
// inline void copy_back(void *dest, const void *src, size_t len) {
//   auto d = reinterpret_cast<char *>(dest);
//   auto s = reinterpret_cast<const char *>(src);
//   d += len - 1;
//   s += len - 1;

//   while (len--) {
//     *(d--) = *(s--);
//   }
// }

// 单个key进行比较。
inline int ix_compare(const char *a, const char *b, ColType type, int col_len) {
  Value left;
  Value right;
  left.type = type;
  right.type = type;
  left.set_data(a, col_len);
  right.set_data(b, col_len);

  return left.compare_value(right);
  // switch (type) {
  //   case TYPE_INT: {
  //     left.set_data(a);
  //   }
  //   case TYPE_FLOAT: {
  //     float fa = *(float *)a;
  //     float fb = *(float *)b;
  //     return (fa < fb) ? -1 : ((fa > fb) ? 1 : 0);
  //   }
  //   case TYPE_STRING: {
  //     Value left;
  //     left.type = TYPE_STRING;
  //     left.set_data(a, col_len);
  //     Value right;
  //     right.type = TYPE_STRING;
  //     right.set_data(b, col_len);
  //     return left.compare_value(right);
  //   }
  //   default:
  //     throw InternalError("Unexpected data type");
  // }
}

// 多个key进行比较。
inline int ix_compare(const char *a, const char *b, const std::vector<ColType> &col_types,
                      const std::vector<int> &col_lens) {
  int offset = 0;
  for (size_t i = 0; i < col_types.size(); ++i) {
    int res = ix_compare(a + offset, b + offset, col_types[i], col_lens[i]);
    if (res != 0) return res;
    offset += col_lens[i];
  }
  return 0;
}

/* 管理B+树中的每个节点 */
class IxNodeHandle {  // 节点管理/
  friend class IxIndexHandle;
  friend class IxScan;

 private:
  const IxFileHdr *file_hdr;  // 节点所在文件的头部信息
  Page *page;                 // 存储节点的页面
  IxPageHdr *page_hdr;  // page->data的第一部分，指针指向首地址，长度为sizeof(IxPageHdr)
  char *
      keys;  // page->data的第二部分，指针指向首地址，长度为file_hdr->keys_size，每个key的长度为file_hdr->col_len
  Rid *rids;  // page->data的第三部分，指针指向首地址

 public:
  IxNodeHandle() = default;

  IxNodeHandle(const IxFileHdr *file_hdr_, Page *page_) : file_hdr(file_hdr_), page(page_) {
    page_hdr = reinterpret_cast<IxPageHdr *>(page->get_data());
    keys = page->get_data() + sizeof(IxPageHdr);
    rids = reinterpret_cast<Rid *>(keys + file_hdr->keys_size_);
  }

  static void show_node(const IxNodeHandle *node);  // debug 用
  static void show_key(const char *key, Rid *rid);  // debug 用

  int get_size() const { return page_hdr->num_key; }

  void set_size(int size) { page_hdr->num_key = size; }

  // #ifdef DEBUG
  // int get_max_size() const { return 5; }  // debug时缩小每页的存放的key数量，方便debug
  // #else
  int get_max_size() const { return file_hdr->btree_order_ + 1; }
  // #endif

  int get_min_size() const { return get_max_size() / 2; }

  int key_at(int i) { return *(int *)get_key(i); }

  /* 得到第i个孩子结点的page_no */
  page_id_t value_at(int i) { return get_rid(i)->page_no; }

  page_id_t get_page_no() const { return page->get_page_id().page_no; }

  PageId get_page_id() const { return page->get_page_id(); }

  page_id_t get_next_leaf() { return page_hdr->next_leaf; }

  page_id_t get_prev_leaf() { return page_hdr->prev_leaf; }

  page_id_t get_parent_page_no() const { return page_hdr->parent; }

  bool is_leaf_page() const { return page_hdr->is_leaf; }

  bool is_root_page() { return get_parent_page_no() == INVALID_PAGE_ID; }

  void set_next_leaf(page_id_t page_no) { page_hdr->next_leaf = page_no; }

  void set_prev_leaf(page_id_t page_no) { page_hdr->prev_leaf = page_no; }

  void set_parent_page_no(page_id_t parent) { page_hdr->parent = parent; }

  char *get_key(int key_idx) const { return keys + key_idx * file_hdr->col_tot_len_; }

  Rid *get_rid(int rid_idx) const { return &rids[rid_idx]; }

  void set_key(int key_idx, const char *key) {
    memcpy(keys + key_idx * file_hdr->col_tot_len_, key, file_hdr->col_tot_len_);
  }

  void set_rid(int rid_idx, const Rid &rid) { rids[rid_idx] = rid; }

  int lower_bound(const char *target) const;

  int upper_bound(const char *target) const;

  void insert_pairs(int pos, const char *key, const Rid *rid, int n);

  page_id_t internal_lookup(const char *key);

  bool leaf_lookup(const char *key, Rid **value);

  int insert(const char *key, const Rid &value);

  // 用于在结点中的指定位置插入单个键值对
  void insert_pair(int pos, const char *key, const Rid &rid) { insert_pairs(pos, key, &rid, 1); }

  void erase_pair(int pos);

  int remove(const char *key);

  /**
   * @brief used in internal node to remove the last key in root node, and return the last child
   *
   * @return the last child
   */
  page_id_t remove_and_return_only_child() {
    assert(get_size() == 1);
    page_id_t child_page_no = value_at(0);
    erase_pair(0);
    assert(get_size() == 0);
    return child_page_no;
  }

  /**
   * @brief 由parent调用，寻找child，返回child在parent中的rid_idx∈[0,page_hdr->num_key)
   * @param child
   * @return int
   */
  int find_child(IxNodeHandle *child) {
    int rid_idx;
    for (rid_idx = 0; rid_idx < page_hdr->num_key; rid_idx++) {
      if (get_rid(rid_idx)->page_no == child->get_page_no()) {
        break;
      }
    }
    assert(rid_idx < page_hdr->num_key);
    return rid_idx;
  }
};

/* B+树 */
class IxIndexHandle {
  friend class IxScan;
  friend class IxManager;

 private:
  DiskManager *disk_manager_;
  BufferPoolManager *buffer_pool_manager_;
  int fd_;  // 存储B+树的文件
  IxFileHdr
      *file_hdr_;  // 存了root_page，但其初始化为2（第0页存FILE_HDR_PAGE，第1页存LEAF_HEADER_PAGE）
                   // 因为分裂策略和合并策略，可以保证LEAF_HEADER_PAGE始终为1
  std::shared_mutex root_latch_;

 public:
  IxIndexHandle(DiskManager *disk_manager, BufferPoolManager *buffer_pool_manager, int fd);

  const std::vector<ColType> &get_index_col_types() const { return file_hdr_->col_types_; }
  const std::vector<int> &get_index_col_lens() const { return file_hdr_->col_lens_; }
  // for search
  bool get_value(const char *key, std::vector<Rid> *result, Transaction *transaction);

  page_id_t insert_entry(const char *key, const Rid &value, Transaction *transaction);

  bool delete_entry(const char *key, Transaction *transaction);

  Iid lower_bound(const char *key);

  Iid upper_bound(const char *key);

  Iid leaf_end() const;

  Iid leaf_begin() const;

  int get_fd() const { return fd_; }

  // #ifdef DEBUG
  //  public:
  //   void show_tree();
  // #else
 private:
  void show_tree();  // debug 用
                     // #endif

 private:
  // 下面所有函数，页面由谁fetch，则由谁unpin。如果页面作为参数传递，则函数内部不会进行释放。
  // 如果页面需要离开函数继续存在，则以返回值形式返回。

  std::unique_ptr<IxNodeHandle> find_leaf_page(const char *key, Operation operation,
                                               Transaction *transaction, bool find_first = false);

  // for insert

  IxNodeHandle *split(IxNodeHandle *node);

  void insert_into_parent(IxNodeHandle *old_node, const char *key, IxNodeHandle *new_node,
                          Transaction *transaction);

  // for delete

  bool coalesce_or_redistribute(IxNodeHandle *node, Transaction *transaction = nullptr,
                                bool *root_is_latched = nullptr);
  bool adjust_root(IxNodeHandle *old_root_node);

  void redistribute(IxNodeHandle *neighbor_node, IxNodeHandle *node, IxNodeHandle *parent,
                    int index);

  bool coalesce(IxNodeHandle **neighbor_node, IxNodeHandle **node, IxNodeHandle **parent, int index,
                Transaction *transaction, bool *root_is_latched);

 private:
  // 辅助函数
  void update_root_page_no(page_id_t root) { file_hdr_->root_page_ = root; }

  bool is_empty() const { return file_hdr_->root_page_ == IX_NO_PAGE; }

  // for get/create node
  IxNodeHandle *fetch_node(int page_no) const;

  IxNodeHandle *create_node();

  // for maintain data structure
  void maintain_parent(IxNodeHandle *node);

  void erase_leaf(IxNodeHandle *leaf);

  void release_node_handle(IxNodeHandle &node);

  void maintain_child(IxNodeHandle *node, int child_idx);

  // for index test
  Rid get_rid(const Iid &iid) const;
};