/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "planner.h"

#include <memory>

#include "common/common.h"
#include "common/context.h"
#include "errors.h"
#include "execution/executor_delete.h"
#include "execution/executor_index_scan.h"
#include "execution/executor_insert.h"
#include "execution/executor_nestedloop_join.h"
#include "execution/executor_projection.h"
#include "execution/executor_seq_scan.h"
#include "execution/executor_update.h"
#include "index/ix.h"
#include "optimizer/optimizer.h"
#include "optimizer/plan.h"
#include "parser/ast.h"
#include "record_printer.h"
#include "system/sm_meta.h"

bool Planner::get_index_cols(std::string tab_name, std::vector<Condition> curr_conds,
                             std::vector<std::string> &index_col_names) {
  TabMeta &tab = sm_manager_->db_.get_table(tab_name);

  // 匹配策略：只要条件中有一个条件匹配到索引的字段就算找到索引。所以匹配到的索引和索引创建的时间有关。
  // 但是不管了，匹配到就完事了。
  for (auto &index : tab.indexes) {
    index_col_names.clear();

    for (auto &col : index.cols) {
      auto iter =
          // 使用索引条件 ： 首先最左匹配，最左匹配的时候最左的比较不能为 <> 也不能是 IN。
          std::find_if(curr_conds.begin(), curr_conds.end(), [&col, &tab_name](Condition &cond) {
            if ((cond.is_rhs_val || cond.is_sub_query) && cond.lhs_col.tab_name == tab_name &&
                cond.op != OP_NE && cond.op != OP_IN && cond.op != OP_NOT_IN &&
                cond.lhs_col.col_name == col.name) {
              return true;
            }

            // 条件为等值，右表为列，左右列表名不同。
            // 这是归并连接的条件，可以使用索引。
            if (cond.op == OP_EQ && cond.is_rhs_column &&
                cond.rhs_col.tab_name != cond.lhs_col.tab_name) {
              return true;
            }

            return false;
          });

      if (iter == curr_conds.end()) {
        break;
      }

      // 归并连接的条件可以用来判断是否可以使用索引，但是不能作为索引的筛选条件。
      index_col_names.push_back(col.name);
    }

    if (!index_col_names.empty()) {
      return true;
    }
  }

  return false;
}

/**
 * @brief 表算子条件谓词生成
 *
 * @param conds 条件
 * @param tab_names 表名
 * @return std::vector<Condition>
 */
std::vector<Condition> pop_conds(std::vector<Condition> &conds, std::string tab_names) {
  // auto has_tab = [&](const std::string &tab_name) {
  //     return std::find(tab_names.begin(), tab_names.end(), tab_name) != tab_names.end();
  // };
  std::vector<Condition> solved_conds;
  auto it = conds.begin();
  while (it != conds.end()) {
    // 右边为常量，左表名与参数相等，可作为该表扫描算子的筛选条件，。
    if ((it->is_rhs_val && tab_names.compare(it->lhs_col.tab_name) == 0) ||
        // 右边为列，并且左右表名相等，均为参数传入的表名，可以作为该表扫描算子的筛选条件。
        (it->is_rhs_column && tab_names == it->lhs_col.tab_name &&
         it->lhs_col.tab_name.compare(it->rhs_col.tab_name) == 0) ||
        // 下面两个没什么好说的。
        (tab_names.compare(it->lhs_col.tab_name) == 0 && it->is_sub_query) ||
        (tab_names.compare(it->lhs_col.tab_name) == 0 && it->is_value_list)) {
      solved_conds.emplace_back(std::move(*it));
      it = conds.erase(it);
    } else if (it->is_rhs_column && it->op == OP_EQ &&
               (tab_names == it->lhs_col.tab_name || tab_names == it->rhs_col.tab_name) &&
               it->lhs_col.tab_name.compare(it->rhs_col.tab_name) != 0) {
      // 右边为列，条件为等值，表名与左表相等（其实这个左右无所谓），并且左右表名不相等。可以认为是一个归并连接条件。
      // 但是这是一个join的条件，不应该直接弹出。
      solved_conds.push_back(*it);
      ++it;
    } else {
      it++;
    }
  }
  return solved_conds;
}

int push_conds(Condition *cond, std::shared_ptr<Plan> plan) {
  if (auto x = std::dynamic_pointer_cast<ScanPlan>(plan)) {
    if (x->tab_name_.compare(cond->lhs_col.tab_name) == 0) {
      return 1;
    } else if (x->tab_name_.compare(cond->rhs_col.tab_name) == 0) {
      return 2;
    } else {
      return 0;
    }
  } else if (auto x = std::dynamic_pointer_cast<JoinPlan>(plan)) {
    int left_res = push_conds(cond, x->left_);
    // 条件已经下推到左子节点
    if (left_res == 3) {
      return 3;
    }
    int right_res = push_conds(cond, x->right_);
    // 条件已经下推到右子节点
    if (right_res == 3) {
      return 3;
    }
    // 左子节点或右子节点有一个没有匹配到条件的列
    if (left_res == 0 || right_res == 0) {
      return left_res + right_res;
    }
    // 左子节点匹配到条件的右边
    if (left_res == 2) {
      // 需要将左右两边的条件变换位置
      std::map<CompOp, CompOp> swap_op = {
          {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT},
          {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
      };
      std::swap(cond->lhs_col, cond->rhs_col);
      cond->op = swap_op.at(cond->op);
    }
    x->conds_.emplace_back(std::move(*cond));
    return 3;
  }
  return false;
}

std::shared_ptr<Plan> pop_scan(int *scantbl, std::string table,
                               std::vector<std::string> &joined_tables,
                               std::vector<std::shared_ptr<Plan>> plans) {
  for (size_t i = 0; i < plans.size(); i++) {
    auto x = std::dynamic_pointer_cast<ScanPlan>(plans[i]);
    if (x->tab_name_.compare(table) == 0) {
      scantbl[i] = 1;
      joined_tables.emplace_back(x->tab_name_);
      return plans[i];
    }
  }
  return nullptr;
}

std::shared_ptr<Query> Planner::logical_optimization(std::shared_ptr<Query> query,
                                                     Context *context) {
  // TODO 实现逻辑优化规则

  return query;
}

std::shared_ptr<Plan> Planner::physical_optimization(std::shared_ptr<Query> query,
                                                     Context *context) {
  std::shared_ptr<Plan> plan = make_one_rel(query);

  // 其他物理优化
  // group_by 计划生成。
  plan = generate_aggregation_or_group_plan(query, std::move(plan));

  // 处理orderby
  plan = generate_sort_plan(query, std::move(plan));

  return plan;
}

std::shared_ptr<Plan> Planner::make_one_rel(std::shared_ptr<Query> query) {
  auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);
  std::vector<std::string> tables = query->tables;
  // // Scan table , 生成表算子列表tab_nodes
  bool need_sleep = false;
  std::vector<std::shared_ptr<Plan>> table_scan_executors(tables.size());
  for (size_t i = 0; i < tables.size(); i++) {
    auto curr_conds = pop_conds(query->conds, tables[i]);
    // int index_no = get_indexNo(tables[i], curr_conds);
    std::vector<std::string> index_col_names;
    bool index_exist = get_index_cols(tables[i], curr_conds, index_col_names);
    // bool index_exist = false;
    need_sleep = !index_exist;
    // index_col_names.clear();
    // table_scan_executors[i] = std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, tables[i],
    //                                                      curr_conds, index_col_names);

    if (index_exist == false) {  // 该表没有索引
      index_col_names.clear();
      table_scan_executors[i] = std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, tables[i],
                                                           curr_conds, index_col_names);
    } else {  // 存在索引
      table_scan_executors[i] = std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, tables[i],
                                                           curr_conds, index_col_names);
    }
  }

  // 只有一个表，不需要join。
  if (tables.size() == 1) {
    return table_scan_executors[0];
  }
  // 获取where条件
  auto conds = std::move(query->conds);
  std::shared_ptr<Plan> table_join_executors;

  int scantbl[tables.size()];
  for (size_t i = 0; i < tables.size(); i++) {
    scantbl[i] = -1;
  }
  // 假设在ast中已经添加了jointree，这里需要修改的逻辑是，先处理jointree，然后再考虑剩下的部分
  if (conds.size() >= 1) {
    // 有连接条件

    // 根据连接条件，生成第一层join
    std::vector<std::string> joined_tables(tables.size());
    auto it = conds.begin();
    while (it != conds.end()) {
      std::shared_ptr<Plan> left, right;
      left = pop_scan(scantbl, it->lhs_col.tab_name, joined_tables, table_scan_executors);
      right = pop_scan(scantbl, it->rhs_col.tab_name, joined_tables, table_scan_executors);
      std::vector<Condition> join_conds{*it};
      std::string left_tbl_name = it->lhs_col.tab_name;
      std::string right_tbl_name = tables[0];
      // bool test_compare = left_tbl_name == right_tbl_name;
      // int test_compare2 = left_tbl_name.compare(right_tbl_name);
      bool need_reverse = left_tbl_name == right_tbl_name;
      // 建立join
      //  判断使用哪种join方式
      if (enable_nestedloop_join && enable_sortmerge_join) {
        // 默认nested loop join
        table_join_executors = std::make_shared<JoinPlan>(
            T_NestLoop, std::move(left), std::move(right), join_conds, need_reverse);
      } else if (enable_nestedloop_join) {
        table_join_executors = std::make_shared<JoinPlan>(
            T_NestLoop, std::move(left), std::move(right), join_conds, need_reverse);
      } else if (enable_sortmerge_join) {
        table_join_executors = std::make_shared<JoinPlan>(
            T_SortMerge, std::move(left), std::move(right), join_conds, need_reverse, need_sleep);
      } else {
        // error
        throw RMDBError("No join executor selected!");
      }

      // table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(left),
      // std::move(right), join_conds);
      it = conds.erase(it);
      break;
    }

    // 根据连接条件，生成第2-n层join
    it = conds.begin();
    while (it != conds.end()) {
      std::shared_ptr<Plan> left_need_to_join_executors = nullptr;
      std::shared_ptr<Plan> right_need_to_join_executors = nullptr;
      bool isneedreverse = false;
      if (std::find(joined_tables.begin(), joined_tables.end(), it->lhs_col.tab_name) ==
          joined_tables.end()) {
        left_need_to_join_executors =
            pop_scan(scantbl, it->lhs_col.tab_name, joined_tables, table_scan_executors);
      }
      if (std::find(joined_tables.begin(), joined_tables.end(), it->rhs_col.tab_name) ==
          joined_tables.end()) {
        right_need_to_join_executors =
            pop_scan(scantbl, it->rhs_col.tab_name, joined_tables, table_scan_executors);
        isneedreverse = true;
      }

      if (left_need_to_join_executors != nullptr && right_need_to_join_executors != nullptr) {
        std::vector<Condition> join_conds{*it};
        std::shared_ptr<Plan> temp_join_executors =
            std::make_shared<JoinPlan>(T_NestLoop, std::move(left_need_to_join_executors),
                                       std::move(right_need_to_join_executors), join_conds, false);
        table_join_executors = std::make_shared<JoinPlan>(
            T_NestLoop, std::move(temp_join_executors), std::move(table_join_executors),
            std::vector<Condition>(), false);
      } else if (left_need_to_join_executors != nullptr ||
                 right_need_to_join_executors != nullptr) {
        if (isneedreverse) {
          std::map<CompOp, CompOp> swap_op = {
              {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT},
              {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
          };
          std::swap(it->lhs_col, it->rhs_col);
          it->op = swap_op.at(it->op);
          left_need_to_join_executors = std::move(right_need_to_join_executors);
        }
        std::vector<Condition> join_conds{*it};
        table_join_executors =
            std::make_shared<JoinPlan>(T_NestLoop, std::move(left_need_to_join_executors),
                                       std::move(table_join_executors), join_conds, false);
      } else {
        push_conds(std::move(&(*it)), table_join_executors);
      }
      it = conds.erase(it);
    }
  } else {
    table_join_executors = table_scan_executors[0];
    scantbl[0] = 1;
  }

  // 连接剩余表
  for (size_t i = 0; i < tables.size(); i++) {
    if (scantbl[i] == -1) {
      table_join_executors = std::make_shared<JoinPlan>(
          T_NestLoop, std::move(table_scan_executors[i]), std::move(table_join_executors),
          std::vector<Condition>(), false);
    }
  }

  return table_join_executors;
}

std::shared_ptr<Plan> Planner::generate_aggregation_or_group_plan(std::shared_ptr<Query> query,
                                                                  std::shared_ptr<Plan> plan) {
  auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);

  if (!x->has_aggregation && !x->has_group) {
    return plan;
  }

  return std::make_shared<AggregationAndGroupPlan>(
      T_AggregationAndGroup, std::move(plan), query->group_cols, query->cols, query->having_conds);
}

std::shared_ptr<Plan> Planner::generate_sort_plan(std::shared_ptr<Query> query,
                                                  std::shared_ptr<Plan> plan) {
  auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);
  if (!x->has_sort) {
    return plan;
  }

  return std::make_shared<SortPlan>(T_Sort, std::move(plan), query->sort_col,
                                    x->order->orderby_dir == ast::OrderBy_DESC);
}

/**
 * @brief select plan 生成
 *
 * @param sel_cols select plan 选取的列
 * @param tab_names select plan 目标的表
 * @param conds select plan 选取条件
 */
std::shared_ptr<Plan> Planner::generate_select_plan(std::shared_ptr<Query> query,
                                                    Context *context) {
  // 逻辑优化
  query = logical_optimization(std::move(query), context);

  // 物理优化
  auto sel_cols = query->cols;
  std::shared_ptr<Plan> plannerRoot = physical_optimization(query, context);

  plannerRoot =
      std::make_shared<ProjectionPlan>(T_Projection, std::move(plannerRoot), std::move(sel_cols));

  return plannerRoot;
}

bool Planner::check_sub_query(std::vector<Condition> &conditions, Context *context,
                              const std::string &tab_name) {
  bool duplicate = false;
  for (auto &cond : conditions) {
    if (cond.is_sub_query) {
      cond.sub_query_plan = do_planner(cond.analyzed_sub_query, context);
    }
  }

  return duplicate;
}

// 生成DDL语句和DML语句的查询执行计划
std::shared_ptr<Plan> Planner::do_planner(std::shared_ptr<Query> query, Context *context) {
  std::shared_ptr<Plan> plannerRoot;
  if (auto x = std::dynamic_pointer_cast<ast::CreateTable>(query->parse)) {
    // create table;
    std::vector<ColDef> col_defs;
    for (auto &field : x->fields) {
      if (auto sv_col_def = std::dynamic_pointer_cast<ast::ColDef>(field)) {
        ColDef col_def = {.name = sv_col_def->col_name,
                          .type = interp_sv_type(sv_col_def->type_len->type),
                          .len = sv_col_def->type_len->len};
        col_defs.push_back(col_def);
      } else {
        throw InternalError("Unexpected field type");
      }
    }
    plannerRoot =
        std::make_shared<DDLPlan>(T_CreateTable, x->tab_name, std::vector<std::string>(), col_defs);
  } else if (auto x = std::dynamic_pointer_cast<ast::DropTable>(query->parse)) {
    // drop table;
    plannerRoot = std::make_shared<DDLPlan>(T_DropTable, x->tab_name, std::vector<std::string>(),
                                            std::vector<ColDef>());
  } else if (auto x = std::dynamic_pointer_cast<ast::CreateIndex>(query->parse)) {
    // create index;
    plannerRoot =
        std::make_shared<DDLPlan>(T_CreateIndex, x->tab_name, x->col_names, std::vector<ColDef>());
  } else if (auto x = std::dynamic_pointer_cast<ast::DropIndex>(query->parse)) {
    // drop index
    plannerRoot =
        std::make_shared<DDLPlan>(T_DropIndex, x->tab_name, x->col_names, std::vector<ColDef>());
  } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(query->parse)) {
    // insert;
    plannerRoot =
        std::make_shared<DMLPlan>(T_Insert, std::shared_ptr<Plan>(), x->tab_name, query->values,
                                  std::vector<Condition>(), std::vector<SetClause>());
  } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(query->parse)) {
    // delete;
    // 生成表扫描方式
    std::shared_ptr<Plan> table_scan_executors;
    // 只有一张表，不需要进行物理优化了
    // int index_no = get_indexNo(x->tab_name, query->conds);
    std::vector<std::string> index_col_names;

    if (check_sub_query(query->conds, context, x->tab_name)) {
      // throw InternalError("delete table in sub query tables");
    }

    bool index_exist = get_index_cols(x->tab_name, query->conds, index_col_names);

    if (index_exist == false) {  // 该表没有索引
      index_col_names.clear();
      table_scan_executors = std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name,
                                                        query->conds, index_col_names);
    } else {  // 存在索引
      table_scan_executors = std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name,
                                                        query->conds, index_col_names);
    }

    plannerRoot =
        std::make_shared<DMLPlan>(T_Delete, table_scan_executors, x->tab_name, std::vector<Value>(),
                                  query->conds, std::vector<SetClause>());
  } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(query->parse)) {
    // update;
    // 生成表扫描方式
    std::shared_ptr<Plan> table_scan_executors;
    // 只有一张表，不需要进行物理优化了
    // int index_no = get_indexNo(x->tab_name, query->conds);
    std::vector<std::string> index_col_names;
    if (check_sub_query(query->conds, context, x->tab_name)) {
      //   throw InternalError("update table in sub query tables");
    }

    bool index_exist = get_index_cols(x->tab_name, query->conds, index_col_names);

    if (index_exist == false) {  // 该表没有索引
      index_col_names.clear();
      table_scan_executors = std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name,
                                                        query->conds, index_col_names);
    } else {  // 存在索引
      table_scan_executors = std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name,
                                                        query->conds, index_col_names);
    }
    plannerRoot = std::make_shared<DMLPlan>(T_Update, table_scan_executors, x->tab_name,
                                            std::vector<Value>(), query->conds, query->set_clauses);
  } else if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse)) {
    std::shared_ptr<plannerInfo> root = std::make_shared<plannerInfo>(x);

    // 生成select语句的查询执行计划
    check_sub_query(query->conds, context, "");

    std::shared_ptr<Plan> projection = generate_select_plan(std::move(query), context);

    plannerRoot =
        std::make_shared<DMLPlan>(T_select, projection, std::string(), std::vector<Value>(),
                                  std::vector<Condition>(), std::vector<SetClause>());
  } else {
    throw InternalError("Unexpected AST root");
  }
  return plannerRoot;
}