/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "log_recovery.h"

#include <cstddef>
#include <cstdint>
#include <memory>

#include "common/config.h"
#include "common/context.h"
#include "common/log.h"
#include "index/ix_index_handle.h"
#include "record/rm_defs.h"
#include "record/rm_scan.h"
#include "recovery/log_defs.h"
#include "recovery/log_manager.h"
#include "storage/page.h"
#include "system/sm_meta.h"
#include "transaction/transaction.h"

/**
 * @description: analyze阶段，需要获得脏页表（DPT）和未完成的事务列表（ATT）
 */
void RecoveryManager::analyze() {
  LogRecord log_record;
  auto log_data = buffer_.buffer_;
  int &offset = buffer_.offset_;
  int file_offset = 0;

  int persis = 0;

  {  // 读的当前的flushed_lsn.
    int master_fd = disk_manager_->open_file(LOG_MASTER_FILE_NAME);
    disk_manager_->read_page(master_fd, HEADER_PAGE_ID, (char *)&persis, sizeof(lsn_t));
    disk_manager_->close_file(master_fd);
  }

  while (disk_manager_->read_log(log_data, LOG_BUFFER_SIZE, file_offset) > 0) {
    offset = 0;

    while (!buffer_.is_full(LOG_HEADER_SIZE)) {
      log_record.deserialize(log_data + offset);

      if (buffer_.is_full(log_record.log_tot_len_) || log_record.lsn_ <= 0) {
        break;
      }

      file_offset += log_record.log_tot_len_;

      std::unique_ptr<LogRecord> log;

      switch (log_record.log_type_) {
        case LogType::ABORT: {
          log.reset(new AbortLogRecord);
          log->deserialize(log_data + offset);
          offset += log->log_tot_len_;
          ATT_table_.erase(log->log_tid_);
        } break;
        case LogType::BEGIN: {
          log.reset(new BeginLogRecord);
          log->deserialize(log_data + offset);
          offset += log->log_tot_len_;
          ATT_table_[log->log_tid_] = {TxnStatus::Running, log->lsn_, false};
        } break;
        case LogType::COMMIT: {
          log.reset(new CommitLogRecord);
          log->deserialize(log_data + offset);
          offset += log->log_tot_len_;
          ATT_table_.erase(log->log_tid_);
        } break;
        case LogType::DELETE: {
          log.reset(new DeleteLogRecord);
          log->deserialize(log_data + offset);
          offset += log->log_tot_len_;
          auto delete_log = static_cast<DeleteLogRecord *>(log.get());

          std::string tab_name(delete_log->table_name_, delete_log->table_name_size_);
          table_names_.emplace(tab_name);

          ATT_table_[log->log_tid_].last_lsn = log->lsn_;

          auto tab = sm_manager_->fhs_[tab_name].get();
          PageId page_id{tab->GetFd(), delete_log->rid_.page_no};
          Page *page = buffer_pool_manager_->fetch_page(page_id);

          if (page->get_page_lsn() < delete_log->lsn_) {
            auto &redo = DPT_table_[page_id];
            redo.redo_logs_.push_back(delete_log->lsn_);
            redo.table_file_ = tab;
          }

          buffer_pool_manager_->unpin_page(page_id, false);
        } break;
        case LogType::INSERT: {
          log.reset(new InsertLogRecord);
          log->deserialize(log_data + offset);
          offset += log->log_tot_len_;

          auto insert_log = static_cast<InsertLogRecord *>(log.get());

          std::string tab_name(insert_log->table_name_, insert_log->table_name_size_);

          table_names_.emplace(tab_name);

          ATT_table_[log->log_tid_].last_lsn = log->lsn_;

          auto tab = sm_manager_->fhs_[tab_name].get();
          PageId page_id{tab->GetFd(), insert_log->rid_.page_no};
          Page *page = buffer_pool_manager_->fetch_page(page_id);

          if (page->get_page_lsn() < insert_log->lsn_) {
            auto &redo = DPT_table_[page_id];
            redo.redo_logs_.push_back(insert_log->lsn_);
            redo.table_file_ = tab;
          }

          buffer_pool_manager_->unpin_page(page_id, false);
        } break;
        case LogType::UPDATE: {
          log.reset(new UpdateLogRecord);
          log->deserialize(log_data + offset);
          offset += log->log_tot_len_;

          auto update_log = static_cast<UpdateLogRecord *>(log.get());

          std::string tab_name(update_log->table_name_, update_log->table_name_size_);

          table_names_.emplace(tab_name);

          ATT_table_[log->log_tid_].last_lsn = log->lsn_;

          auto tab = sm_manager_->fhs_[tab_name].get();
          PageId page_id{tab->GetFd(), update_log->rid_.page_no};
          Page *page = buffer_pool_manager_->fetch_page(page_id);

          if (page->get_page_lsn() < update_log->lsn_) {
            auto &redo = DPT_table_[page_id];
            redo.redo_logs_.push_back(update_log->lsn_);
            redo.table_file_ = tab;
          }

          buffer_pool_manager_->unpin_page(page_id, false);
        } break;
        default: {
        }
      }

      // log->format_print();

      persis = std::max(persis, log->lsn_ + 1);
      log_list_.push_back(log.get());
      log_table_[log->lsn_] = std::move(log);
    }

    memset(log_data, 0, LOG_BUFFER_SIZE);
  }

  log_manager_->persist_lsn_ = persis;
  log_manager_->global_lsn_ = persis;

  for (auto iter = ATT_table_.begin(); iter != ATT_table_.end();) {
    if (iter->second.is_implicit) {
      iter = ATT_table_.erase(iter);
    } else {
      ++iter;
    }
  }
}

/**
 * @description: 重做所有未落盘的操作
 */
void RecoveryManager::redo() {
  Transaction txn(INVALID_TXN_ID);
  Context context(nullptr, nullptr, &txn);

  for (auto &[page_id, redo] : DPT_table_) {
    for (auto &redo_lsn : redo.redo_logs_) {
      auto &log_record = log_table_[redo_lsn];
      txn.txn_id_ = log_record->log_tid_;
      txn.set_prev_lsn(log_record->lsn_);

      // printf("redo : \n");
      // log_record->format_print();
      // printf("\n");

      switch (log_record->log_type_) {
        case LogType::DELETE: {
          auto delete_log = static_cast<DeleteLogRecord *>(log_record.get());
          redo.table_file_->delete_record(delete_log->rid_, &context);
        } break;
        case LogType::INSERT: {
          auto insert_log = static_cast<InsertLogRecord *>(log_record.get());
          redo.table_file_->insert_record(insert_log->rid_, insert_log->insert_value_.data,
                                          &context);
        } break;
        case LogType::UPDATE: {
          auto update_log = static_cast<UpdateLogRecord *>(log_record.get());
          redo.table_file_->update_record(update_log->rid_, update_log->new_value_.data, &context);
        } break;
        default: {
        }
      }
    }
  }

  DPT_table_.clear();
}

/**
 * @description: 回滚未完成的事务
 */
void RecoveryManager::undo() {
  Transaction txn(INVALID_TXN_ID);
  Context context(nullptr, nullptr, &txn);

  for (auto iter = log_list_.rbegin(); iter != log_list_.rend(); ++iter) {
    auto log_record = *iter;
    auto i = ATT_table_.find(log_record->log_tid_);
    if (ATT_table_.end() == i) {
      continue;
    }

    auto &txn_status = i->second;
    txn.txn_id_ = log_record->log_tid_;
    txn.set_prev_lsn(log_record->lsn_);

    // printf("undo : \n");
    // log_record->format_print();
    // printf("\n");

    if (txn_status.txn_status == TxnStatus::Running && log_record->lsn_ == txn_status.last_lsn) {
      txn_status.last_lsn = log_record->prev_lsn_;

      switch (log_record->log_type_) {
        case LogType::DELETE: {
          auto delete_log = static_cast<DeleteLogRecord *>(log_record);
          std::string tab_name(delete_log->table_name_, delete_log->table_name_size_);

          auto tab = sm_manager_->fhs_.at(tab_name).get();

          tab->insert_record(delete_log->rid_, delete_log->delete_value_.data, &context);
        } break;
        case LogType::UPDATE: {
          auto update_log = static_cast<UpdateLogRecord *>(log_record);
          std::string tab_name(update_log->table_name_, update_log->table_name_size_);

          auto tab = sm_manager_->fhs_.at(tab_name).get();

          tab->update_record(update_log->rid_, update_log->old_value_.data, &context);

        } break;
        case LogType::INSERT: {
          auto insert_log = static_cast<InsertLogRecord *>(log_record);

          std::string tab_name(insert_log->table_name_, insert_log->table_name_size_);

          auto tab = sm_manager_->fhs_.at(tab_name).get();

          tab->delete_record(insert_log->rid_, &context);
        } break;
        default: {
        }
      }
    }
  }

  // 索引重建。
  for (auto &tab_name : table_names_) {
    auto &tab_meta = sm_manager_->db_.get_table(tab_name);

    if (tab_meta.indexes.empty()) {
      continue;
    }

    std::vector<std::vector<std::string>> ix_names;
    ix_names.reserve(tab_meta.indexes.size());

    // 保存该表的索引信息。
    for (auto &ix_meta : tab_meta.indexes) {
      ix_names.emplace_back(std::vector<std::string>());
      ix_names.back().reserve(ix_meta.cols.size());

      for (auto &ix_col : ix_meta.cols) {
        ix_names.back().emplace_back(ix_col.name);
      }
    }

    // 对索引删除重建。
    for (auto &ix_name : ix_names) {
      sm_manager_->drop_index(tab_name, ix_name, nullptr);
      sm_manager_->create_index(tab_name, ix_name, nullptr);
    }
  }

  table_names_.clear();

  buffer_pool_manager_->flush_all_pages();

  // 每次恢复之后，日志文件就没用了，可以直接删掉。
  disk_manager_->destroy_log();
  disk_manager_->create_file(LOG_FILE_NAME);

  log_manager_->begin_flush_loop();
  log_list_.clear();
  log_table_.clear();
  ATT_table_.clear();
}