/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "lock_manager.h"

#include <readline/readline.h>

#include <algorithm>
#include <memory>
#include <mutex>

#include "common/config.h"
#include "common/log.h"
#include "parser/ast.h"
#include "transaction/txn_defs.h"

// 间隙锁插入检查
bool LockManager::check_gap_lock(LockRequestQueue& lock_queue, txn_id_t t_id, LockMode lock_mode,
                                 std::string col_name, Value left_val, Value right_val,
                                 bool star_wait_die) {
  std::unordered_map<std::string, lockGapList> gap_list_;
  if (lock_mode == LockMode::Index_GAP_X) {
    // left_val = right_val
    gap_list_ = lock_queue.gap_list_s;
    for (auto& gap_loc_list : gap_list_[col_name].request_gap_queue_) {
      if (gap_loc_list->txn_id_ != t_id) {
        if ((left_val.compare_value(gap_loc_list->low_key_) == CompResult::LT ||
             left_val.compare_value(gap_loc_list->upper_key_) == CompResult::GT)) {
          continue;
        } else if ((left_val.compare_value(gap_loc_list->low_key_) == CompResult::EQ) &&
                   gap_loc_list->low_key_.is_closed == false) {
          continue;
        } else if (left_val.compare_value(gap_loc_list->upper_key_) == CompResult::EQ &&
                   gap_loc_list->upper_key_.is_closed == false) {
          continue;
        } else {
          // 如何区间重叠且 当前插入的事务时间戳更小,且 该次检查考虑时间戳则允许等待
          if (gap_loc_list->txn_id_ > t_id && star_wait_die)
            continue;
          else
            return false;
        }
      }
    }
  } else if (lock_mode == LockMode::INDEX_GAP_S) {
    gap_list_ = lock_queue.gap_list_x;
    // 对于 插入的间隙队列 ，low_key_ = upper_key_
    for (auto& gap_loc_list : gap_list_[col_name].request_gap_queue_) {
      if (gap_loc_list->txn_id_ != t_id) {
        if ((gap_loc_list->low_key_.compare_value(left_val) == CompResult::LT ||
             gap_loc_list->low_key_.compare_value(right_val) == CompResult::GT)) {
          continue;
        } else if ((gap_loc_list->low_key_.compare_value(left_val) == CompResult::EQ) &&
                   left_val.is_closed == false) {
          continue;
        } else if (gap_loc_list->low_key_.compare_value(right_val) == CompResult::EQ &&
                   right_val.is_closed == false) {
          continue;
        } else {
          if (gap_loc_list->txn_id_ > t_id && star_wait_die)
            continue;
          else
            return false;
        }
      }
    }
  }
  return true;
}
bool LockManager::grant_lock(LockRequestQueue& lock_queue, txn_id_t t_id, LockMode lock_mode,
                             LockRequest& lock_request) {
  for (auto& i : lock_queue.locked_queue_) {
    // ASSERT(i->granted_, "locked queue should be granted.");
    if (!check_lock(i->lock_mode_, lock_mode)) {
      return false;
    }
  }

  // 现在锁和占用资源的事务的锁是相容的，或者现在资源没有被占用。
  // 因为可以保证队列的时间戳为降序，因此需要检查前面部分。

  // 是否和前面部分兼容，如果兼容，那么这些部分都可以获取锁。

  auto i = lock_queue.request_queue_.begin();

  for (; (*i)->txn_id_ != t_id; ++i) {
    // ASSERT(!(*i)->granted_, "request queue should be not granted.");
    if (!check_lock((*i)->lock_mode_, lock_mode)) {
      return false;
    }
  }

  // ASSERT(i != lock_queue.request_queue_.end(), "grant lock found failed.");

  (*i)->granted_ = true;
  lock_queue.locked_queue_.push_back(*i);
  lock_queue.request_queue_.erase(i);
  // 可以
  return true;
}
// 间隙锁wait-die策略
bool LockManager::check_gap_lock_timep(LockRequestQueue& lock_queue, txn_id_t t_id,
                                       LockMode lock_mode) {
  // [X(1)] S(2),S(3) -> 这样的允许等待，S(3)不会被abort。
  for (auto& it : lock_queue.request_queue_) {
    if (t_id > it->txn_id_ && (it->lock_mode_ == LockMode::INDEX_GAP_S)) {
      return false;
    }
  }

  for (auto& it : lock_queue.locked_queue_) {
    if (t_id > it->txn_id_) {
      return false;
    }
  }

  return true;
}
bool LockManager::check_timep(LockRequestQueue& lock_queue, txn_id_t t_id, LockMode lock_mode) {
  // [X(1)] S(2),S(3) -> 这样的允许等待，S(3)不会被abort。
  for (auto& it : lock_queue.request_queue_) {
    if (t_id > it->txn_id_ || check_lock(it->lock_mode_, lock_mode)) {
      return false;
    }
  }

  // [X(1)] S(2),S(3) -> 这样的不允许等待，S(3)会被abort。
  // for (auto& it : lock_queue.request_queue_) {
  //   if (t_id > it->txn_id_) {
  //     return false;
  //   }
  // }

  for (auto& it : lock_queue.locked_queue_) {
    if (t_id > it->txn_id_) {
      return false;
    }
  }

  return true;
}

bool LockManager::check_lock(LockMode left_mode, LockMode right_mode) {
  // 间隙锁和别的锁都兼容，但是间隙锁和间隙锁的话则需要符合
  // if (left_mode == LockMode::INDEX_GAP || right_mode == LockMode::INDEX_GAP) return true;
  // X，不兼容。
  if (left_mode == LockMode::EXLUCSIVE) {
    return false;
  }

  // SIX只和IS兼容。
  if (left_mode == LockMode::S_IX && right_mode != LockMode::INTENTION_SHARED) {
    return false;
  }

  // S只和S与IS兼容。
  if (left_mode == LockMode::SHARED && right_mode != LockMode::SHARED &&
      right_mode != LockMode::INTENTION_SHARED) {
    return false;
  }

  // IX之和IX与IS兼容。
  if (left_mode == LockMode::INTENTION_EXCLUSIVE && right_mode != LockMode::INTENTION_EXCLUSIVE &&
      right_mode != LockMode::INTENTION_SHARED) {
    return false;
  }

  // IS不和X兼容。
  if (left_mode == LockMode::INTENTION_SHARED && right_mode == LockMode::EXLUCSIVE) {
    return false;
  }

  return true;
}

bool LockManager::is_upgrade(LockMode lock_mode1, LockMode lock_mode2) {
  bool is = (lock_mode1 == LockMode::INTENTION_SHARED);
  bool s = (lock_mode1 == LockMode::SHARED);
  bool ix = (lock_mode1 == LockMode::INTENTION_EXCLUSIVE);
  bool six = (lock_mode1 == LockMode::S_IX);

  if ((!is || lock_mode2 == LockMode::INTENTION_SHARED) &&  // IS锁除了本身，都可以申请。
      (!s || (lock_mode2 != LockMode::EXLUCSIVE &&
              lock_mode2 != LockMode::S_IX)) &&  // S锁可以申请X锁或者SIX锁
      (!ix || (lock_mode2 != LockMode::EXLUCSIVE &&
               lock_mode2 != LockMode::S_IX)) &&  // IX锁可以申请X锁或者SIX锁。
      (!six || lock_mode2 != LockMode::EXLUCSIVE)) {
    return false;
  }

  return true;
}

bool LockManager::lock(Transaction* txn, LockDataId lock_data, LockMode lock_mode) {
  TransactionState state = txn->get_state();
  txn_id_t t_id = txn->get_transaction_id();

  // 第一步，检查事务状态.
  if (state == TransactionState::ABORTED || state == TransactionState::COMMITTED) {
    txn->set_state(TransactionState::ABORTED);
    throw TransactionAbortException(t_id, AbortReason::LOCK_IN_COMMIT_OR_ABORT);
    return false;
  }

  if (state == TransactionState::SHRINKING) {
    txn->set_state(TransactionState::ABORTED);
    throw TransactionAbortException(t_id, AbortReason::LOCK_ON_SHIRINKING);
    return false;
  }

  latch_.lock();

  auto& que = lock_table_[lock_data];
  std::unique_lock<std::mutex> lock(que.latch_);

  latch_.unlock();

  // 第二步，检查是否是锁升级。
  {
    auto iter =
        std::find_if(que.locked_queue_.begin(), que.locked_queue_.end(),
                     [t_id](std::shared_ptr<LockRequest>& lr) { return lr->txn_id_ == t_id; });

    // 这个地方事务肯定一定获得了锁。
    // 原因：如果一个事务没获得这个锁，要么在等待，要么已经abort。不可能去申请锁升级。
    if (iter != que.locked_queue_.end()) {
      // ASSERT((*iter)->granted_, "not grant lock");  // 必然已经获得。

      if ((*iter)->lock_mode_ == lock_mode) {
        txn->set_state(TransactionState::GROWING);
        return true;
      }

      // 我的理解：
      // 已经申请到一个高级锁，那么就不需要继续申请低级锁。直接返回即可.
      if (!is_upgrade((*iter)->lock_mode_, lock_mode)) {
        txn->set_state(TransactionState::GROWING);
        return true;
      }

      // 该锁升级了。
      que.locked_queue_.erase(iter);
      txn->get_lock_set()->erase(lock_data);  // 忘记删除原来的了。
    }
  }

  // 第三步，对兼容锁特判处理。
  // 当锁兼容的时候，不需要等待？直接进行操作？
  // 评测机说就是这么做。
  {
    bool f = true;
    for (auto& lock_request : que.locked_queue_) {
      if (!check_lock(lock_request->lock_mode_, lock_mode)) {
        f = false;
        break;
      }
    }

    if (f) {
      std::shared_ptr<LockRequest> lock_request(new LockRequest(t_id, lock_mode));
      lock_request->granted_ = true;
      que.locked_queue_.push_back(lock_request);
      txn->get_lock_set()->insert(lock_data);
      txn->set_state(TransactionState::GROWING);
      return true;
    }
  }

  // 如果满足这个条件，则可以等待，可以等待，则一定可以最终获取锁。
  // 因此grant_lock中不需要对事务做abort处理。
  if (!check_timep(que, t_id, lock_mode)) {
    txn->set_state(TransactionState::ABORTED);
    throw TransactionAbortException(t_id, AbortReason::DEADLOCK_PREVENTION);
    return false;
  }

  std::shared_ptr<LockRequest> lock_request(new LockRequest(t_id, lock_mode));
  que.request_queue_.push_back(lock_request);

  while (!grant_lock(que, t_id, lock_mode, *lock_request)) {
    que.cv_.wait(lock);
  }

  // 事务给与锁集。
  txn->get_lock_set()->insert(lock_data);
  txn->set_state(TransactionState::GROWING);  // 设置状态，其实默认设置成growing不得了？

  return true;
}

// 间隙锁
bool LockManager::lock_gap(Transaction* txn, LockDataId lock_data, LockMode lock_mode,
                           std::vector<std::string> col_names, std::vector<Value> insert_values) {
  TransactionState state = txn->get_state();
  txn_id_t t_id = txn->get_transaction_id();

  // 第一步，检查事务状态.
  if (state == TransactionState::ABORTED || state == TransactionState::COMMITTED) {
    txn->set_state(TransactionState::ABORTED);
    throw TransactionAbortException(t_id, AbortReason::LOCK_IN_COMMIT_OR_ABORT);
    return false;
  }

  if (state == TransactionState::SHRINKING) {
    txn->set_state(TransactionState::ABORTED);
    throw TransactionAbortException(t_id, AbortReason::LOCK_ON_SHIRINKING);
    return false;
  }

  latch_.lock();

  auto& que = lock_table_[lock_data];
  std::unique_lock<std::mutex> lock(que.latch_);

  latch_.unlock();

  /* 检测是否符合间隙锁*/
  // 如果不符合间隙锁，则如果满足wait-die 则等待而不是abort
  size_t i = 0;
  for (auto& col_name : col_names) {
    Value insert_val = insert_values[i++];
    // 如果有重叠，且wait-die策略使用后也无法等待着则直接abort
    if (check_gap_lock(que, t_id, lock_mode, col_name, insert_val, insert_val, true) == false) {
      txn->set_state(TransactionState::ABORTED);
      throw TransactionAbortException(t_id, AbortReason::DEADLOCK_PREVENTION);
      return false;
    }
    // if (!check_gap_lock_timep(que, t_id, lock_mode)) {
    //   txn->set_state(TransactionState::ABORTED);
    //   throw TransactionAbortException(t_id, AbortReason::DEADLOCK_PREVENTION);
    //   return false;
    // }
    else {
      // std::shared_ptr<LockRequest> lock_request(new LockRequest(t_id, lock_mode));
      // que.request_queue_.push_back(lock_request);
      // 否则，检测一下；如果不重叠则不需要等待；如果重叠，则需要等待
      while (!check_gap_lock(que, t_id, lock_mode, col_name, insert_val, insert_val, false)) {
        que.cv_.wait(lock);
      }
      // 如果等到了直接删除等待，不用放到申请到锁队列中
      // auto iter = std::find_if(que.request_queue_.begin(), que.request_queue_.end(),
      //                          [t_id, lock_mode](std::shared_ptr<LockRequest>& lr) {
      //                            return t_id == lr->txn_id_ && lr->lock_mode_ == lock_mode;
      //                          });
      // que.request_queue_.erase(iter);
    }
    que.is_gap_lock_ = true;
    que.txn_gap_cols[t_id].push_back(col_name);
    std::shared_ptr<gapLockRequest> lock_gap_request(
        new gapLockRequest(t_id, LockMode::Index_GAP_X, insert_val, insert_val));
    que.gap_list_x[col_name].request_gap_queue_.push_back(lock_gap_request);
    txn->get_lock_set()->insert(lock_data);  // 用于commit 之后释放间隙锁
  }
  /* wait-die check_gap_lock_timep*/
  // if (!check_timep(que, t_id, lock_mode)) {
  //   txn->set_state(TransactionState::ABORTED);
  //   throw TransactionAbortException(t_id, AbortReason::DEADLOCK_PREVENTION);
  //   return false;
  // }
  return true;
}

/**
 * @description: 申请行级共享锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID 记录所在的表的fd
 * @param {int} tab_fd
 */
bool LockManager::lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
  LockDataId lock_data_id(tab_fd, rid, LockDataType::RECORD);
  return lock(txn, lock_data_id, LockMode::SHARED);
}

/**
 * @description: 申请行级排他锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID
 * @param {int} tab_fd 记录所在的表的fd
 */
bool LockManager::lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
  LockDataId lock_data_id(tab_fd, rid, LockDataType::RECORD);
  return lock(txn, lock_data_id, LockMode::EXLUCSIVE);
}

/**
 * @description: 申请间隙锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 * @param {int} index_fd 目标索引的fd
 */
bool LockManager::lock_gap_on_index(Transaction* txn, int tab_fd, int index_fd,
                                    std::vector<std::string> col_names,
                                    std::vector<Value> insert_values) {
  LockDataId lock_data_id(tab_fd, index_fd, LockDataType::RECORD);
  return lock_gap(txn, lock_data_id, LockMode::Index_GAP_X, col_names, insert_values);
}
/**
 * @description: 添加间隙锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 * @param {int} index_fd 目标索引的fd
 */
bool LockManager::add_index_gap_lock(Transaction* txn, int tab_fd, int index_fd,
                                     std::vector<std::string> index_col_names_,
                                     std::unordered_map<std::string, Value> left_,
                                     std::unordered_map<std::string, Value> right_) {
  LockDataId lock_data_id(tab_fd, index_fd, LockDataType::RECORD);
  TransactionState state = txn->get_state();
  txn_id_t t_id = txn->get_transaction_id();

  // 第一步，检查事务状态.
  if (state == TransactionState::ABORTED || state == TransactionState::COMMITTED) {
    txn->set_state(TransactionState::ABORTED);
    throw TransactionAbortException(t_id, AbortReason::LOCK_IN_COMMIT_OR_ABORT);
    return false;
  }

  if (state == TransactionState::SHRINKING) {
    txn->set_state(TransactionState::ABORTED);
    throw TransactionAbortException(t_id, AbortReason::LOCK_ON_SHIRINKING);
    return false;
  }

  latch_.lock();

  auto& que = lock_table_[lock_data_id];
  std::unique_lock<std::mutex> lock(que.latch_);
  latch_.unlock();
  for (auto& index_col_name : index_col_names_) {
    Value cur_left_value, cur_right_value;
    cur_left_value.set_min(), cur_right_value.set_max();
    // 遍历左边界
    for (const auto& pair : left_) {
      std::string index_col = pair.first;
      Value index_value = pair.second;
      if (index_col_name == index_col) {
        cur_left_value = index_value;
        break;
      }
    }
    // 遍历右边界
    for (const auto& pair : right_) {
      std::string index_col = pair.first;
      Value index_value = pair.second;
      if (index_col_name == index_col) {
        cur_right_value = index_value;
        break;
      }
    }
    // 如果两个都是原始值，说明这个列where时候没有受到限制
    if (cur_left_value.is_min() && cur_right_value.is_max()) continue;
    if (check_gap_lock(que, t_id, LockMode::INDEX_GAP_S, index_col_name, cur_left_value,
                       cur_right_value, true) == false) {
      txn->set_state(TransactionState::ABORTED);
      throw TransactionAbortException(t_id, AbortReason::DEADLOCK_PREVENTION);
      return false;
    } else {
      while (!check_gap_lock(que, t_id, LockMode::INDEX_GAP_S, index_col_name, cur_left_value,
                             cur_right_value, false)) {
        que.cv_.wait(lock);
      }
    }
    std::shared_ptr<gapLockRequest> lock_gap_request(
        new gapLockRequest(t_id, LockMode::INDEX_GAP_S, cur_left_value, cur_right_value));
    que.gap_list_s[index_col_name].request_gap_queue_.push_back(lock_gap_request);
    que.is_gap_lock_ = true;
    que.txn_gap_cols[t_id].push_back(index_col_name);  // 当前事务中对哪些索引列加gap锁
    // while (!grant_lock(que, t_id, lock_mode, *lock_request)) {
    //   que.cv_.wait(lock);
    // }
    // 事务给与锁集。
    txn->get_lock_set()->insert(lock_data_id);  // 用于commit 之后释放间隙锁
    txn->set_state(TransactionState::GROWING);  // 设置状态，其实默认设置成growing不得了？
  }
  return true;
}

/**
 * @description: 申请表级读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_shared_on_table(Transaction* txn, int tab_fd) {
  LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
  return lock(txn, lock_data_id, LockMode::SHARED);
}

/**
 * @description: 申请表级写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_exclusive_on_table(Transaction* txn, int tab_fd) {
  LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
  return lock(txn, lock_data_id, LockMode::EXLUCSIVE);
}

/**
 * @description: 申请表级意向读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IS_on_table(Transaction* txn, int tab_fd) {
  LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
  return lock(txn, lock_data_id, LockMode::INTENTION_SHARED);
}

/**
 * @description: 申请表级意向写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IX_on_table(Transaction* txn, int tab_fd) {
  LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
  return lock(txn, lock_data_id, LockMode::INTENTION_EXCLUSIVE);
}

/**
 * @description: 释放锁
 * @return {bool} 返回解锁是否成功
 * @param {Transaction*} txn 要释放锁的事务对象指针
 * @param {LockDataId} lock_data_id 要释放的锁ID
 */
bool LockManager::unlock(Transaction* txn, LockDataId lock_data_id) {
  latch_.lock();

  // ASSERT(lock_table_.count(lock_data_id) > 0, "unlock resource not exists.");

  auto& que = lock_table_[lock_data_id];

  std::unique_lock<std::mutex> lock(que.latch_);

  latch_.unlock();

  txn_id_t t_id = txn->get_transaction_id();
  auto lock_set = txn->get_lock_set();
  // auto state = txn->get_state();

  // ASSERT(lock_set->count(lock_data_id) > 0, "the txn does not have lock in the resource.");
  // ASSERT(txn->get_state() != TransactionState::COMMITTED, "the txn unlock in invalid state.");

  // ASSERT(iter != que.locked_queue_.end(), "lock queue don't have the resource.");
  // ASSERT((*iter)->granted_, "unlock a not granted resource.");

  // lock_set->erase(lock_data_id);
  // 如果是间隙锁
  if (que.is_gap_lock_) {
    std::vector<std::string>& gap_cols = que.txn_gap_cols[t_id];
    for (auto& col : gap_cols) {
      //  lockGapList txn_col_gap_list = que.gap_list_[col];
      // 读间隙锁
      auto iter_gap_s =
          std::find_if(que.gap_list_s[col].request_gap_queue_.begin(),
                       que.gap_list_s[col].request_gap_queue_.end(),
                       [t_id](std::shared_ptr<gapLockRequest>& lr) { return t_id == lr->txn_id_; });
      while (iter_gap_s != que.gap_list_s[col].request_gap_queue_.end()) {
        que.gap_list_s[col].request_gap_queue_.erase(iter_gap_s);
        iter_gap_s = std::find_if(
            que.gap_list_s[col].request_gap_queue_.begin(),
            que.gap_list_s[col].request_gap_queue_.end(),
            [t_id](std::shared_ptr<gapLockRequest>& lr) { return t_id == lr->txn_id_; });
      }
      // 写间隙锁
      auto iter_gap_x =
          std::find_if(que.gap_list_x[col].request_gap_queue_.begin(),
                       que.gap_list_x[col].request_gap_queue_.end(),
                       [t_id](std::shared_ptr<gapLockRequest>& lr) { return t_id == lr->txn_id_; });
      while (iter_gap_x != que.gap_list_x[col].request_gap_queue_.end()) {
        que.gap_list_x[col].request_gap_queue_.erase(iter_gap_x);
        iter_gap_x = std::find_if(
            que.gap_list_x[col].request_gap_queue_.begin(),
            que.gap_list_x[col].request_gap_queue_.end(),
            [t_id](std::shared_ptr<gapLockRequest>& lr) { return t_id == lr->txn_id_; });
      }
    }
  } else {
    auto iter =
        std::find_if(que.locked_queue_.begin(), que.locked_queue_.end(),
                     [t_id](std::shared_ptr<LockRequest>& lr) { return t_id == lr->txn_id_; });

    if (iter != que.locked_queue_.end()) {
      que.locked_queue_.erase(iter);
    }
  }

  txn->set_state(TransactionState::SHRINKING);

  que.cv_.notify_all();
  return true;
}